################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import Dict, List, Set, Tuple, Union, Optional

import torch
from torch import Tensor
import torch.nn.functional as F
from ase.data import (  # reference_states, # stored lattice constant, bond length, etc.
    atomic_masses,
    atomic_names,
    atomic_numbers,
    chemical_symbols,
    covalent_radii,
    ground_state_magnetic_moments,
    vdw_radii,
)

ATOMIC_SYMBOLS: List[str]
ATOMIC_NUMBERS: Dict[str, int]
ATOMIC_NAMES: List[str]
ATOMIC_MASSES: List[float]
ATOMIC_MAGMOM: List[float]
ATOMIC_VDW_RADII: List[float]
ATOMIC_COVALENT_RADII: List[float]

ATOMIC_SYMBOLS = chemical_symbols
ATOMIC_NUMBERS = atomic_numbers
ATOMIC_NAMES = atomic_names
ATOMIC_NAMES[0] = "Dummy"
ATOMIC_MASSES = atomic_masses.tolist()
ATOMIC_VDW_RADII = vdw_radii.tolist()
ATOMIC_COVALENT_RADII = (
    covalent_radii.tolist()
)  # if covalent_radii = 0.2, that means this element missing covalent_radii in ase
ATOMIC_MAGMOM = ground_state_magnetic_moments.tolist()  # ground state


class Element:
    def __init__(self, zs: Union[List[int], Tuple[int, ...]]):
        if isinstance(zs, tuple):
            self.zs = list(zs)
        elif isinstance(zs, list):
            self.zs = zs
        else:
            raise TypeError(f"{zs} must be a list or tuple of integers.")
        if not all(isinstance(num, int) for num in self.zs):
            raise ValueError(f"All elements in {zs} must be integers.")
        self.atomic_numbers = sorted(zs)
        self.num_elements = len(zs)
        self.z_to_idx = {z: i for i, z in enumerate(self.atomic_numbers)}

        # # Torch
        # self.lookup_table = torch.full((max(self.atomic_numbers) + 1,), -1, dtype=torch.int64)
        # for idx, z in enumerate(self.atomic_numbers):
        #     self.lookup_table[z] = idx

    def z2idx(self, z: int) -> int:
        return self.z_to_idx[z]

    # def torch_z2idx(self, z: torch.Tensor) -> torch.Tensor:
    #     return self.lookup_table[z]

    # def torch_z2onehot(self, zs: torch.Tensor) -> torch.Tensor:
    #     """
    #     :param zs: shape (N,),
    #     :return: shape (N, num_classes), one-hot
    #     """
    #     idxs = self.lookup_table[zs]  # shape (N,)
    #     return F.one_hot(idxs, num_classes=self.num_elements)

    def idx2symbol(self, idx: int) -> int:
        z = self.idx2z(idx)
        return ATOMIC_SYMBOLS[z]

    def idx2mass(self, idx: int) -> float:
        z = self.idx2z(idx)
        return ATOMIC_MASSES[z]

    def idx2name(self, idx: int) -> str:
        z = self.idx2z(idx)
        return ATOMIC_NAMES[z]

    def idx2covalent_radii(self, idx: int) -> float:
        z = self.idx2z(idx)
        return ATOMIC_COVALENT_RADII[z]

    def idx2vdw_radii(self, idx: int) -> float:
        z = self.idx2z(idx)
        return ATOMIC_VDW_RADII[z]

    def idx2magmom(self, idx: int) -> float:
        z = self.idx2z(idx)
        return ATOMIC_MAGMOM[z]

    def __len__(self) -> int:
        return len(self.atomic_numbers)

    def __repr__(self):
        return f"Element in the model: {(z for z in self.atomic_numbers)}"


class TorchElement(Element):
    def __init__(self, zs: Union[List[int], Tuple[int, ...]]):

        super().__init__(zs)

        self.lookup_table = torch.full(
            (max(self.atomic_numbers) + 1,), -1, dtype=torch.int64
        )
        for idx, z in enumerate(self.atomic_numbers):
            self.lookup_table[z] = idx

    def z2idx(self, z: Tensor) -> Tensor:
        return self.lookup_table[z]

    def z2onehot(self, zs: Tensor) -> Tensor:
        """
        :param zs: shape (N,),
        :return: shape (N, num_classes), one-hot
        """
        lookup_table = self.lookup_table.to(zs.device)
        idxs = lookup_table[zs]  # shape (N,)
        return F.one_hot(idxs, num_classes=self.num_elements)


def build_element_lookup(atomic_numbers: Set[int]) -> TorchElement:
    element = TorchElement(sorted(list(atomic_numbers)))
    return element



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import logging
from typing import Dict
from pathlib import Path
from datetime import datetime


import lightning as L
from lightning.pytorch.callbacks import ModelCheckpoint
from hydra.utils import instantiate
from torch_geometric.loader import DataLoader
from pytorch_lightning.callbacks import StochasticWeightAveraging


from .lit_model import LightningWrapperModel
from ..utils.callbacks import PrintMetricsCallback
from ..utils.utils import log_parameters


def build_trainer(cfg: Dict, dataloader_valid: DataLoader = None) -> L.Trainer:
    """Build and configure a PyTorch Lightning Trainer.

    Args:
        cfg: Hydra configuration object containing:
            - logger: Configuration for experiment logger
            - callbacks: Dictionary of callback configurations
            - trainer: Base Trainer configuration

    Returns:
        Fully configured PyTorch Lightning Trainer instance

    Raises:
        ValueError: For missing critical configurations
        RuntimeError: For initialization failures in components
    """

    # === Configuration Preprocessing ===
    if not isinstance(cfg, Dict):
        raise TypeError(f"Expected DictConfig type, got {type(cfg)}")

    # === Logger Configuration ===
    logger_cfg = cfg.get("logger", None)
    logger_instance = None
    if logger_cfg is not None:
        try:
            logger_instance = instantiate(logger_cfg)
        except Exception as e:
            raise RuntimeError(f"Logger initialization failed: {str(e)}") from e

    # === User Callbacks Configuration ===
    initialized_callbacks = []
    for cb_name, cb_config in cfg.get("callbacks", {}).items():
        try:
            if cb_config.get("_target_") is None:
                logging.warning(f"Skipping unconfigured callback: {cb_name}")
                continue

            callback = instantiate({k: v for k, v in cb_config.items() if k != "extra"})
            initialized_callbacks.append(callback)
            logging.debug(
                f"Successfully loaded callback: {cb_name} ({type(callback).__name__})"
            )
        except Exception as e:
            error_msg = (
                f"Callback '{cb_name}' initialization failed\n"
                f"Config: {cb_config}\n"
                f"Error: {str(e)}"
            )
            logging.error(error_msg, exc_info=True)

    # === Built-in callbacks ===
    initialized_callbacks += [PrintMetricsCallback()]

    # === Disallow SWA callback ===
    swa_cbs = [
        cb for cb in initialized_callbacks
        if isinstance(cb, StochasticWeightAveraging)
    ]
    if swa_cbs:
        raise RuntimeError(
            "StochasticWeightAveraging (SWA) is currently not allowed in TACE when used "
            "directly via PyTorch Lightning callbacks.\n\n"
            "Reason: During testing, we observed some strange issues "
            "when using Lightning's built-in SWA.\n\n"
            "Recommended alternative:\n"
            "- Save a series of model checkpoints during training.\n"
            "- After training, use the provided `tace-average` command to compute an averaged "
            "model from any set of models with the same architecture.\n\n"
            "This approach is especially useful in the late stage of training, "
            "particularly when increasing the energy loss weight, and can significantly "
            "improve the model's generalization performance."
        )

    # === Put checkpoint at the end ===
    other_cbs = [
        cb for cb in initialized_callbacks if not isinstance(cb, ModelCheckpoint)
    ]
    checkpoint_cbs = [
        cb for cb in initialized_callbacks if isinstance(cb, ModelCheckpoint)
    ]
    initialized_callbacks = other_cbs + checkpoint_cbs     
    if checkpoint_cbs:
        for cb in checkpoint_cbs:
            logging.info(f"Model checkpoints will be saved to: {cb.dirpath}")
            if not cb.monitor:
                logging.warning(f"Checkpoint in {cb.dirpath} has no monitor metric specified")
    else:
        raise RuntimeError(
            "No ModelCheckpoint callback configured. "
            "You must provide at least one ModelCheckpoint to save the model checkpoints."
        )
    
    try:
        trainer_cfg = cfg["trainer"]
        filter_trainer_cfg = {}
        for k, v in trainer_cfg.items():
            filter_keys = ["logger", "callbacks"]
            if k not in filter_keys:
                filter_trainer_cfg.update({k: v})
        num_batch = 2 if not cfg.get("dataset", {}).get("no_valid_set", False) else 0
        filter_trainer_cfg.update({"num_sanity_val_steps": num_batch})
        trainer = instantiate(
            filter_trainer_cfg,
            logger=logger_instance,
            callbacks=initialized_callbacks,
            # _convert_="partial" # none, partial, all
        )
    except Exception as e:
        error_detail = (
            "Trainer initialization failed\n"
            f"Configuration: {(cfg["trainer"])}\n"
            f"Error: {str(e)}"
        )
        raise RuntimeError(error_detail) from e
    
    logging.info(f"Callbacks: {list(cfg.get("callbacks", {}))}")

    return trainer


# === Training entrypoint ===
def train(
    cfg: Dict,
    statistics,
    model,
    datamodule,
    target_property,
    embedding_property: list[str] = []
):
    lit_model = LightningWrapperModel(
        cfg, 
        model,
        target_property, 
        embedding_property,
        statistics, 
    )

    # Trainer
    trainer = build_trainer(cfg)

    # TRAIN AND VALID
    resume_ckpt = cfg.get("resume_from_model", None)
    if resume_ckpt:
        ckpt_path = Path(resume_ckpt)
        if not ckpt_path.is_file():
            raise FileNotFoundError(f"Checkpoint not exists: {ckpt_path}")
        logging.info(f"Resume Training from checkpoint: {resume_ckpt}")
        trainer.fit(
            lit_model,
            datamodule=datamodule,
            ckpt_path=resume_ckpt,
        )
    else:
        logging.info(lit_model)
        log_parameters(model)
        trainer.fit(
            lit_model,
            datamodule=datamodule,
        )
    
    # TEST
    if cfg['dataset'].get('test_files', None):
        trainer.test(
            lit_model,
            datamodule=datamodule,
            ckpt_path="best",
            verbose=False,
        )
    logging.info("Training completed at %s", datetime.now().strftime("%Y-%m-%d %H:%M"))




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from scipy.constants import (
    pi,
    e,
    c,
    Planck,
    hbar,
    Boltzmann,
    epsilon_0,
    Avogadro,
    R,
    eV,
    angstrom,
    femto,
    pico,
    nano,
    bar,
    calorie,
)

kcalPerMol = 1000 * calorie / Avogadro


# === unit convert ===
kcalpermol2ev = kcalPerMol / eV
ev2kcalpermol = eV / kcalPerMol

pa2eva3 = (1 / eV) / (1e30)
kbar2eva3 = 1e8 * pa2eva3


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import logging
from typing import Dict


import yaml
from importlib.resources import files


from tace import utils


def deep_update(d, u):
    for k, v in u.items():
        if isinstance(v, dict) and isinstance(d.get(k), dict):
            deep_update(d[k], v)
        else:
            d[k] = v
    return d


def update_cfg(cfg: Dict):
    # === load default cfg ===
    config_yaml = files(utils) / "config.yaml"

    with config_yaml.open("r") as f:
        default_cfg = yaml.safe_load(f)

    # update
    ignore_default_config = cfg.get("ignore_default_config", False)

    if not ignore_default_config:
        logging.info("Useing user's config and default config")
        cfg = deep_update(default_cfg, cfg)
    else:
        logging.info("Useing only user's config")

    return cfg



import torch
from cartnn.o3 import Irreps, CartesianHarmonics
from cartnn.util.jit import compile

lmax=3
cart = CartesianHarmonics(Irreps.cartesian_harmonics(lmax, p=-1), normalize=True)
model = compile(cart)
print(model)
v = torch.randn(1, 3)
v = torch.nn.functional.normalize(v, dim=-1)
print(model(v).shape)




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################


import argparse

import torch

from ..lightning import load_tace
from ..lightning.lora import from_lora_to_merged_model
from ..utils._global import DTYPE

ALLOWED_TYPE = ["merged_lora"]

def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-m", "--model",
        type=str,
        required=True,
        help="Model path",
    )
    parser.add_argument(
        "--dtype",
        type=str,
        help="Model dtype",
        choices=['float32', 'float64'],
        default=None,
    )
    parser.add_argument(
        "-t", "--type", 
        type=str, 
        default="merge_lora",
        choices=ALLOWED_TYPE, 
        help="Specify convert type"
    )
    parser.add_argument(
        "-d", "--debug", 
        type=int, 
        default=0,
        help="print some extra information for debug"
    )
    return parser.parse_args()


def count_parameters(model: torch.nn.Module) -> int:
    return sum(p.numel() for p in model.parameters())

def main():
    args = parse_args()
    model = load_tace(args.model, 'cpu', strict=True, use_ema=True)
    if bool(args.debug):
        print(model)
    model_dtype = model.readout_fn.cutoff.dtype
    args_dtype = DTYPE[args.dtype] or model_dtype
    if args_dtype != model_dtype:
        print(f"[Warning] Model dtype does not match args.dtype. Forcing dtype from {model_dtype} to {args_dtype}")
    model.to(dtype=args_dtype)
    if args.type == "merge_lora":
        total_before = count_parameters(model)
        model = from_lora_to_merged_model(model)
        total_after = count_parameters(model)
        if bool(args.debug):
            print(model)
        print("The number of parameters: ")
        print(f"  Your LoRA:     {total_before - total_after}")
        print(f"  Before merged: {total_before}")
        print(f"  After merged:  {total_after}")
        model.to(dtype=args_dtype)
        torch.save(model, args.model + "-merged_lora.pt")
    else:
        raise ValueError(f"Unsupported convert type '{args.type}'. One of {ALLOWED_TYPE} is available.")

if __name__ == "__main__":
    main()


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
"""
Hessians related will be refacored, training is not supported now
"""

from typing import Dict, List, Optional, Tuple


import torch


from ...utils.torch_scatter import scatter_sum


def compute_symmetric_displacement(
        data: Dict[str, torch.Tensor], num_graphs: int
    ) -> torch.Tensor:
    
    displacement = torch.zeros(
        (num_graphs, 3, 3),
        dtype=data["positions"].dtype,
        device=data["positions"].device,
    )
    displacement.requires_grad_(True)
    symmetric_displacement = 0.5 * (displacement + displacement.transpose(-1, -2))

    positions = data["positions"]
    positions.requires_grad_(True)
    if data["lattice"] is None:
        data["lattice"] = torch.zeros(
            num_graphs * 3,
            3,
            dtype=data["positions"].dtype,
            device=data["positions"].device,
        )

    data["positions"] = positions + torch.einsum(
        "be,bec->bc", positions, symmetric_displacement[data["batch"]]
    )

    lattice = data["lattice"]
    data["lattice"] = lattice + torch.matmul(lattice, symmetric_displacement)

    return displacement


def compute_atomic_virials_stresses(
    edge_vector: torch.Tensor,
    edge_forces: torch.Tensor,
    edge_index: torch.Tensor, 
    lattice: torch.Tensor, 
    batch: torch.Tensor,
    num_nodes: torch.Tensor,
    compute_atomic_virials: bool,
    compute_atomic_stresses: bool,
) -> Tuple[torch.Tensor, torch.Tensor]:
    
    atomic_virials = None
    atomic_stresses = None

    if compute_atomic_virials or compute_atomic_stresses:
        edge_virials = torch.einsum("zi,zj->zij", edge_forces, edge_vector)
        atomic_virials_source = scatter_sum(
            edge_virials, edge_index[0], dim=0, dim_size=num_nodes
        )
        atomic_virials_target = scatter_sum(
            edge_virials, edge_index[1], dim=0, dim_size=num_nodes
        )
        atomic_virials = (atomic_virials_source + atomic_virials_target) / 2
        atomic_virials = -1 * (atomic_virials + atomic_virials.transpose(-1, -2)) / 2

        volume = torch.linalg.det(lattice).abs().unsqueeze(-1)
        atomic_stresses = -1 * atomic_virials / volume[batch].view(-1, 1, 1)
        atomic_stresses = torch.where(
            torch.abs(atomic_stresses) < 1e10, atomic_stresses, torch.zeros_like(atomic_stresses)
        )

    return atomic_virials, atomic_stresses


def compute_hessians_vmap(
    forces: torch.Tensor,
    positions: torch.Tensor,
) -> torch.Tensor:
    forces_flatten = forces.view(-1)
    num_elements = forces_flatten.shape[0]

    def get_vjp(v):
        return torch.autograd.grad(
            -1 * forces_flatten,
            positions,
            v,
            retain_graph=True,
            create_graph=False,
            allow_unused=False,
        )

    I_N = torch.eye(num_elements).to(forces.device)
    try:
        chunk_size = 1 if num_elements < 64 else 16
        gradient = torch.vmap(get_vjp, in_dims=0, out_dims=0, chunk_size=chunk_size)(
            I_N
        )[0]
    except RuntimeError:
        gradient = compute_hessians_loop(forces, positions)
    if gradient is None:
        return torch.zeros((positions.shape[0], forces.shape[0], 3, 3))
    return gradient

def compute_hessians_loop(
    forces: torch.Tensor,
    positions: torch.Tensor,
) -> torch.Tensor:
    hessian = []
    for grad_elem in forces.view(-1):
        hess_row = torch.autograd.grad(
            outputs=[-1 * grad_elem],
            inputs=[positions],
            grad_outputs=torch.ones_like(grad_elem),
            retain_graph=True,
            create_graph=False,
            allow_unused=False,
        )[0]
        hess_row = hess_row.detach()  # this makes it very slow? but needs less memory
        if hess_row is None:
            hessian.append(torch.zeros_like(positions))
        else:
            hessian.append(hess_row)
    hessian = torch.stack(hessian)
    return hessian

def sample_force_components(
    n_atoms: int,
    num_samples: int,
    device: torch.device,
) -> torch.Tensor:
    k = min(num_samples, n_atoms)
    atom_idx = torch.randperm(n_atoms, device=device)[:k]
    xyz_idx = torch.randint(0, 3, (k,), device=device)
    return torch.stack([atom_idx, xyz_idx], dim=1)


def build_grad_outputs(
    samples_per_graph: List[torch.Tensor],
    ptr: torch.Tensor,
    total_atoms: int,
    device: torch.device,
) -> torch.Tensor:
    
    all_samples = []

    # Convert local atom indices to global (batch-level) indices
    for graph_id, samples in enumerate(samples_per_graph):
        offset = ptr[graph_id]
        s = samples.clone()
        s[:, 0] += offset  # local -> global atom index
        all_samples.append(s)

    all_samples = torch.cat(all_samples, dim=0) # 1D, effective global atoms
    K_total = all_samples.shape[0]

    grad_outputs = torch.zeros(
        (K_total, total_atoms, 3),
        device=device,
        dtype=all_samples.dtype,
    )

    grad_outputs[
        torch.arange(K_total, device=device),
        all_samples[:, 0],
        all_samples[:, 1],
    ] = 1.0

    return grad_outputs


def compute_force_jacobian(
    forces: torch.Tensor,
    positions: torch.Tensor,
    grad_outputs: torch.Tensor,
    create_graph: bool = True,
) -> torch.Tensor:
    def single_grad(go):
        return torch.autograd.grad(
            outputs=forces,
            inputs=positions,
            grad_outputs=go,
            retain_graph=True,
            create_graph=create_graph,
        )[0]
    return torch.vmap(single_grad)(grad_outputs)


def split_jacobian_per_graph(
    jac: torch.Tensor,
    samples_per_graph: List[torch.Tensor],
    ptr: torch.Tensor,
) -> List[torch.Tensor]:
    jacs_per_graph = []
    row_offset = 0

    for graph_id, samples in enumerate(samples_per_graph):
        k_g = samples.shape[0]
        start, end = ptr[graph_id], ptr[graph_id + 1]

        jac_graph = jac[row_offset : row_offset + k_g, start:end, :]
        jacs_per_graph.append(jac_graph)

        row_offset += k_g

    return jacs_per_graph

def sample_force_jacobian(
    forces: torch.Tensor,
    positions: torch.Tensor,
    ptr: torch.Tensor,
    num_samples: int = 2,
    create_graph: bool = True,
) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:

    device = forces.device
    num_graphs = ptr.numel() - 1
    total_atoms = forces.shape[0]

    samples_per_graph = []
    for g in range(num_graphs):
        n_atoms = (ptr[g + 1] - ptr[g]).item()
        samples = sample_force_components(
            n_atoms=n_atoms,
            num_samples=num_samples,
            device=device,
        ) # [num_samples, 2]
        samples_per_graph.append(samples)

    grad_outputs = build_grad_outputs(
        samples_per_graph=samples_per_graph,
        ptr=ptr,
        total_atoms=total_atoms,
        device=device,
    )

    jac = compute_force_jacobian(
        forces=forces,
        positions=positions,
        grad_outputs=grad_outputs,
        create_graph=create_graph,
    )

    # 4. Split back per graph
    jacs_per_graph = split_jacobian_per_graph(
        jac=jac,
        samples_per_graph=samples_per_graph,
        ptr=ptr,
    )

    return jacs_per_graph, samples_per_graph



import torch
from cartnn import o3

torch.set_printoptions(precision=4, sci_mode=False)
torch.manual_seed(5)
dtype = torch.set_default_dtype(torch.float64)

import torch
from e3nn import o3
from tqdm import tqdm
from typing import Tuple, List
def ICTD(n_total : int) -> Tuple[List[List[int]], List[torch.Tensor]]:
    n_now = 0
    j_now = 0
    path_list = []
    this_path = []
    this_pathmatrix = o3.wigner_3j(0,0,0)
    pathmatrices_list = []
    # generate paths and path matrices
    def paths_generate(n_now, j_now, this_path, this_pathmatrix, n_total):
        if n_now<=n_total:
            this_path.append(j_now)
            for j in [j_now + 1, j_now, j_now - 1]:
                if not (j_now==0 and (j!=1) ) and n_now+1<=n_total:
                    cgmatrix = o3.wigner_3j(1,j_now,j)
                    this_pathmatrix_ = torch.einsum("abc,dce->dabe",this_pathmatrix,cgmatrix)
                    this_pathmatrix_ = this_pathmatrix_.reshape(cgmatrix.shape[0],-1,cgmatrix.shape[-1])
                    paths_generate(n_now+1,j,this_path.copy(),this_pathmatrix_,n_total)
            if n_now == n_total:
                this_pathmatrix = this_pathmatrix.reshape(-1,this_pathmatrix.shape[-1])
                this_pathmatrix = this_pathmatrix*(1./(this_pathmatrix**2).sum(0)[0]**(0.5)) # normalize
                pathmatrices_list.append(this_pathmatrix)
                path_list.append(this_path)
        return     
    paths_generate(n_now, j_now, this_path, this_pathmatrix, n_total)
    decomp_list = []
    cart2sph_list = []
    for path_matrix in tqdm(pathmatrices_list):
            decomp_list.append(path_matrix@path_matrix.T)
            cart2sph_list.append(path_matrix)
    return path_list, decomp_list, cart2sph_list

L1 = 2
L2 = 2
L3 = L1 + L2

T1 = torch.randn(3**L1) @ ICTD(L1)[1][0]
T2 = torch.randn(3**L2) @ ICTD(L2)[1][0]
T = torch.outer(T1, T2).view(-1)


To_list = []
P1S, D1S, C1S = ICTD(L3)
for P1, D1, C1 in zip(P1S, D1S, C1S):
    To = T @ D1
    To_sph = T @ C1
    print(To_sph)
    To_list.append(To)


# G = torch.zeros(len(To_list), len(To_list))
# for i in range(len(To_list)):
#     for j in range(len(To_list)):
#         G[i, j] = torch.dot(To_list[i], To_list[j])
#         if i == j:
#             print(G[i, j])

mat = torch.stack(To_list)
rank = torch.linalg.matrix_rank(mat)
print(f"Number of linearly independent vectors: {rank}")






################################################################################
# Authors: Zemin Xu 
# License: MIT, see LICENSE.md
################################################################################

from math import sqrt
from typing import Tuple


import torch
import opt_einsum_fx


from .paths import generate_path

PATH_ = 4
BATCH_ = 5
CHANNEL_ = 6

class InterEinsumTC(torch.nn.Module):
    '''Reducible tensor product and tensor contraction'''
    def __init__(self, comb: Tuple) -> None:
        super().__init__()
        l1, l2, _, k = comb
        self.comb = comb
        self.expr = generate_path(*comb, True)
        ctr = torch.fx.symbolic_trace(lambda T1, T2: torch.einsum(self.expr, T1, T2))
        self.ctr = (
            opt_einsum_fx.optimize_einsums_full(
                model=ctr,
                example_inputs=(
                    torch.randn([BATCH_] + [CHANNEL_] + [3] * l1),
                    torch.randn([BATCH_] + [CHANNEL_] + [3] * l2),
                ),
            )
        )
        self.normalizer = 1.0 / sqrt(3**k)

    def forward(self, T1: torch.Tensor, T2: torch.Tensor) -> torch.Tensor:
        return self.ctr(T1, T2) * self.normalizer

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self.expr}, comb={self.comb})"


class ProdEinsumTC(torch.nn.Module):
    '''Reducible tensor product and tensor contraction'''
    def __init__(self, comb: Tuple) -> None:
        super().__init__()
        l1, l2, _, k = comb
        self.comb = comb
        expr = generate_path(*comb, True)
        inputs, output = expr.split("->")
        in1, in2 = [x.strip() for x in inputs.split(",")]
        in1 = "a" + in1
        output = "a" + output
        self.expr = in1 + "," + in2 + "->" + output
        ctr = torch.fx.symbolic_trace(lambda T1, T2: torch.einsum(self.expr, T1, T2))
        self.ctr = (
            opt_einsum_fx.optimize_einsums_full(
                model=ctr,
                example_inputs=(
                    torch.randn([PATH_] + [BATCH_] + [CHANNEL_] + [3] * l1),
                    torch.randn([BATCH_] + [CHANNEL_] + [3] * l2),
                ),
            )
        )
        self.normalizer = 1.0 / sqrt(3**k)

    def forward(self, T1: torch.Tensor, T2: torch.Tensor) -> torch.Tensor:
        return self.ctr(T1, T2) * self.normalizer

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self.expr}, comb={self.comb})"




################################################################################
# Authors: Zemin Xu
# Description: To avoid dependency,
# copy from https://github.com/katsura-jp/pytorch-cosine-annealing-with-warmup
# License: MIT, see LICENSE.md
################################################################################
"""
MIT License

Copyright (c) 2022 Naoki Katsura

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import math
import torch
from torch.optim.lr_scheduler import _LRScheduler


class CosineAnnealingWarmupRestarts(_LRScheduler):
    """
    optimizer (Optimizer): Wrapped optimizer.
    first_cycle_steps (int): First cycle step size.
    cycle_mult(float): Cycle steps magnification. Default: -1.
    max_lr(float): First cycle's max learning rate. Default: 0.1.
    min_lr(float): Min learning rate. Default: 0.001.
    warmup_steps(int): Linear warmup step size. Default: 0.
    gamma(float): Decrease rate of max learning rate by cycle. Default: 1.
    last_epoch (int): The index of last epoch. Default: -1.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        first_cycle_steps: int,
        cycle_mult: float = 1.0,
        max_lr: float = 0.1,
        min_lr: float = 0.001,
        warmup_steps: int = 0,
        gamma: float = 1.0,
        last_epoch: int = -1,
    ):
        assert warmup_steps < first_cycle_steps

        self.first_cycle_steps = first_cycle_steps  # first cycle step size
        self.cycle_mult = cycle_mult  # cycle steps magnification
        self.base_max_lr = max_lr  # first max learning rate
        self.max_lr = max_lr  # max learning rate in the current cycle
        self.min_lr = min_lr  # min learning rate
        self.warmup_steps = warmup_steps  # warmup step size
        self.gamma = gamma  # decrease rate of max learning rate by cycle

        self.cur_cycle_steps = first_cycle_steps  # first cycle step size
        self.cycle = 0  # cycle count
        self.step_in_cycle = last_epoch  # step size of the current cycle

        super(CosineAnnealingWarmupRestarts, self).__init__(optimizer, last_epoch)

        # set learning rate min_lr
        self.init_lr()

    def init_lr(self):
        self.base_lrs = []
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = self.min_lr
            self.base_lrs.append(self.min_lr)

    def get_lr(self):
        if self.step_in_cycle == -1:
            return self.base_lrs
        elif self.step_in_cycle < self.warmup_steps:
            return [
                (self.max_lr - base_lr) * self.step_in_cycle / self.warmup_steps
                + base_lr
                for base_lr in self.base_lrs
            ]
        else:
            return [
                base_lr
                + (self.max_lr - base_lr)
                * (
                    1
                    + math.cos(
                        math.pi
                        * (self.step_in_cycle - self.warmup_steps)
                        / (self.cur_cycle_steps - self.warmup_steps)
                    )
                )
                / 2
                for base_lr in self.base_lrs
            ]

    def step(self, epoch=None):
        if epoch is None:
            epoch = self.last_epoch + 1
            self.step_in_cycle = self.step_in_cycle + 1
            if self.step_in_cycle >= self.cur_cycle_steps:
                self.cycle += 1
                self.step_in_cycle = self.step_in_cycle - self.cur_cycle_steps
                self.cur_cycle_steps = (
                    int((self.cur_cycle_steps - self.warmup_steps) * self.cycle_mult)
                    + self.warmup_steps
                )
        else:
            if epoch >= self.first_cycle_steps:
                if self.cycle_mult == 1.0:
                    self.step_in_cycle = epoch % self.first_cycle_steps
                    self.cycle = epoch // self.first_cycle_steps
                else:
                    n = int(
                        math.log(
                            (
                                epoch / self.first_cycle_steps * (self.cycle_mult - 1)
                                + 1
                            ),
                            self.cycle_mult,
                        )
                    )
                    self.cycle = n
                    self.step_in_cycle = epoch - int(
                        self.first_cycle_steps
                        * (self.cycle_mult**n - 1)
                        / (self.cycle_mult - 1)
                    )
                    self.cur_cycle_steps = self.first_cycle_steps * self.cycle_mult ** (
                        n
                    )
            else:
                self.cur_cycle_steps = self.first_cycle_steps
                self.step_in_cycle = epoch

        self.max_lr = self.base_max_lr * (self.gamma**self.cycle)
        self.last_epoch = math.floor(epoch)
        for param_group, lr in zip(self.optimizer.param_groups, self.get_lr()):
            param_group["lr"] = lr



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
from typing import Dict, List, Any


import torch
from torch import Tensor, nn


from .radial import RadialBasis
from .ch import LegacyCartesianHarmonics2
from .mlp import MLP
from .inter import Interaction
from .prod import SelfContraction
from .embedding import UniversalInvariantEmbedding, UniversalEquivariantEmbedding
from .utils import Graph


class TACEDescriptor(torch.nn.Module):
    def __init__(
        self,
        cutoff: float,
        avg_num_neighbors: int,
        num_layers: int,
        atomic_numbers: List[int],
        Lmax: int,
        lmax: int,
        num_channel: List[int] = 64,
        num_channel_hidden: List[int] = 64,
        bias: bool = False,
        radial_basis: Dict = {},
        angular_basis: Dict = {},
        radial_mlp: Dict = {},
        inter: Dict = {},
        prod: Dict = {},
        universal_embedding: Dict[str, Dict[str, bool | int | float |str]] = {},
        target_irreps: List[str] = [],
        **kwargs,
    ):
        super().__init__()

        # === init ===
        self.invariant_embedding_property = universal_embedding['invariant_embedding_property']
        self.equivariant_embedding_property = universal_embedding['equivariant_embedding_property']
        self.register_buffer("num_layers", torch.tensor(num_layers, dtype=torch.int64))
        self.register_buffer("atomic_numbers", torch.tensor(atomic_numbers, dtype=torch.int64))
        self.register_buffer("cutoff", torch.tensor(cutoff, dtype=torch.get_default_dtype()))

        # input, hiiden, output irreps
        ls_in = []      # in of inter
        ls_hidden = []  # out of inter and in of prod
        ls_out = []     # out of prod, out of sc
        for idx in range(num_layers):
            ls_hidden.append(list(range(lmax[idx] + 1)))
            ls_in.append([0]) if idx == 0 else ls_in.append(list(range(Lmax[idx] + 1)))
            ls_out.append(target_irreps if idx == num_layers - 1 else list(range(Lmax[idx] + 1)))

        # === element embedding ===
        self.node_embedding = MLP(
            len(atomic_numbers),
            num_channel,
            hidden_dim=[],
            act=None,
            bias=False,
            forward_weight_init=True,
            enable_layer_norm=False,
        )

        # === universal embedding ===
        if len(universal_embedding['invariant_embedding_property']) > 0:
            self.uie_embedding = UniversalInvariantEmbedding(
                num_channel,
                universal_embedding.get("invariant", {}),
            )
        if len(universal_embedding['equivariant_embedding_property']) > 0:
            self.uee_embeddings = nn.ModuleList()
            for _ in range(num_layers):
                self.uee_embeddings.append(
                    UniversalEquivariantEmbedding(
                        universal_embedding.get("equivariant", {}),
                        atomic_numbers,
                        num_channel,
                    )
                )

        # === radial basis ===
        self.radial_embedding = RadialBasis(
            cutoff=cutoff,
            num_basis=radial_basis.get('num_radial_basis', 8),
            polynomial_cutoff=radial_basis.get('polynomial_cutoff', 5),
            radial_basis=radial_basis.get('radial_basis', 'j0'),
            distance_transform=radial_basis.get('distance_transform', None),
            order=radial_basis.get('order', 0),
            trainable=radial_basis.get('trainable', False),
            apply_cutoff=radial_basis.get("apply_cutoff", True)
        )

        # === angular basis ===
        self.angular_embedding = LegacyCartesianHarmonics2(
            max(lmax), 
            angular_basis.get('norm', True), 
            angular_basis.get('traceless', True)
        )

        # === Interaction Layer ===
        self.interactions = nn.ModuleList(
            [
                Interaction(
                    idx,
                    num_layers,
                    num_channel,
                    num_channel_hidden,
                    bias,
                    max(ls_in[idx]),
                    max(ls_hidden[idx]),
                    ls_out[idx],
                    atomic_numbers,
                    avg_num_neighbors,
                    self.radial_embedding.out_dim,
                    radial_mlp,
                    inter,
                )
                for idx in range(num_layers)
            ]
        )

        # === Product Layer ===
        self.products = nn.ModuleList(
            [
                SelfContraction(
                    num_channel,
                    num_channel_hidden,
                    max(ls_hidden[idx]),
                    ls_out[idx],
                    atomic_numbers,
                    prod,
                    bias,
                    idx,
                    num_layers,
                )
                for idx in range(num_layers)
            ]
        )
 
    def forward(self, data: Dict[str, Tensor], graph: Graph) -> Dict[str, Any]:

        lmp = graph.lmp
        nlocal, _ = graph.lmp_natoms
  
        # === node initialize (element and uie) ===
        node_feats = {0: self.node_embedding(data['node_attrs'])}
        uie_feats = None
        if hasattr(self, "uie_embedding"):
            uie_data = {}
            for k in self.invariant_embedding_property:
                uie_data.update({k: data[k]})
            uie_feats = self.uie_embedding(data["batch"], uie_data)
            node_feats[0] = node_feats[0] + uie_feats
        if hasattr(self, 'uee_embeddings'):
            uee_data = {}
            for k in self.equivariant_embedding_property:
                    uee_data.update({k: data[k]})

        # === edge initialize (radial and angular) ===
        edge_feats, cutoff = self.radial_embedding(
            graph.edge_length,
            data['node_attrs'],
            data['edge_index'],
            self.atomic_numbers,
        )
        edge_attrs = {}
        normed_edge_vector = graph.edge_vector / graph.edge_length
        edge_attrs = self.angular_embedding(normed_edge_vector)

        # === representation Learning ===
        descriptors = []
        for idx, (inter, prod) in enumerate(zip(self.interactions, self.products)):
            node_attrs_total = data['node_attrs']
            node_attrs_slice = data['node_attrs']
            if lmp and idx > 0:
                node_attrs_slice = node_attrs_slice[:nlocal] 
            node_feats, sc = inter(
                node_feats,
                node_attrs_total, 
                node_attrs_slice, 
                edge_feats, 
                edge_attrs, 
                data['edge_index'],
                cutoff,
                graph,
            )
            if lmp and idx == 0:
                node_attrs_slice = node_attrs_slice[:nlocal] # nlocal
            if hasattr(self, 'uee_embeddings'):
                node_feats = self.uee_embeddings[idx](node_feats, node_attrs_slice, data["batch"], uee_data)
            node_feats = prod(node_feats, node_attrs_slice, sc)
            descriptors.append(node_feats)

        return {
            "uie_feats": uie_feats,
            "descriptors": descriptors,
        }




RADIAL_BASIS = {
    "radial_basis": "j0",
    "num_radial_basis": 8,
    "distance_transform": None,
    "polynomial_cutoff": 5,
    "order": 0,
    "trainable": False,
    "apply_cutoff": True,
}


ANGULAR_BASIS = {
    "traceless": True,
    "norm": True,
}


RADIAL_MLP = {
    "hidden": [64, 64, 64],
    "act": "silu",
    "bias": False,
}


INTER = {
    "l1l2": None,
    "conv_weights": ["edge_ij"],
    "normalizer": "avg_num_neighbors",
    "nonlinearity": {
      "type": None,
      "gate": 'silu', 
    },
    "sc": {
      "use_first_sc": False, 
      "from": "current_message",
    },
    "use_resnet": False,   
}


PROD = {
    "l1l2": None,
    "l3l1": None,
    "correlation": 3,
    "element_aware": True,
    "coupled_channel": True,
}


READOUT_EMLP = {
    "hidden": [16], 
    "act": "silu", 
    "gate": "silu",
    "bias": False,
    "use_all_layer": True,
    "enable_uie_readout": False,
}


SCALE_SHIFT = {
    "scale_type": "rms_forces",
    "shift_type": "mean_delta_energy_per_atom",
    "scale_trainable": False,
    "shift_trainable": False,
}


SHORT_RANGE = {
    "zbl": {
      "enable": False,
      "trainable": False,
    }
}

LONG_RANGE = {
    "les": {
        "enable": False,
        "les_arguments": {
            "n_layers": 3,
            "n_hidden": [32, 16],
            "add_linear_nn": True,
            "output_scaling_factor": 0.1,
            "sigma": 1.0,
            "dl": 2.0,
            "remove_mean": True,
            "epsilon_factor": 1.0,
            "use_atomwise": False,
            "compute_bec": False,
            "bec_output_index": None,
        },
    },
}


CONSERVATION = {
    "charges": {
        "method": "lagrangian",
    },
}


UNIVERSAL_EMBEDDING = {
    "invariant": {
        "level": {
            "enable": False,
            "num_embeddings": -1,
        },
        "spin_multiplicity": {
            "enable": False,
            "num_embeddings": -1,
        },
        "charges": {
            "enable": False,
            "act": "silu",
        },
        "total_charge": {
            "enable": False,
            "act": "silu",
        },
        "initial_collinear_magmoms": {
            "enable": False,
            "act": "silu",
        },
        "temperature": {
            "enable": False,
            "act": "silu",
        },
        "electron_temperature": {
            "enable": False,
            "act": "silu",
        },
    },
    "equivariant": {
        "initial_noncollinear_magmoms": {
            "enable": False,
            "normalizer": 1.0,
        },
        "electric_field": {
            "enable": False,
            "normalizer": 1.0,
        },
        "magnetic_field": {
            "enable": False,
            "normalizer": 1.0,
        },
    },
}


from typing import Dict, Any, List, Tuple

def check_model_config(cfg: Dict[str, Any]):
    assert isinstance(cfg['radial_basis'], Dict), "cfg.model.config.radial_basis must be a Dict"
    assert isinstance(cfg['radial_mlp'], Dict), "cfg.model.config.radial_mlp must be a Dict"
    assert isinstance(cfg['angular_basis'], Dict), "cfg.model.config.angular_basis must be a Dict"
    assert isinstance(cfg['inter'], Dict), "cfg.model.config.inter must be a Dict"
    assert isinstance(cfg['prod'], Dict), "cfg.model.config.prod must be a Dict"
    assert isinstance(cfg['readout_emlp'], Dict), "cfg.model.config.readout_emlp must be a Dict"
    assert isinstance(cfg['scale_shift'], Dict), "cfg.model.config.scale_shift must be a Dict"
    assert cfg['short_range'] is None or isinstance(cfg['short_range'], Dict), "cfg.model.config.short_range must be a Dict or None"
    assert cfg['long_range'] is None or isinstance(cfg['long_range'], Dict), "cfg.model.config.long_range must be a Dict or None"
    assert cfg['embedding_property'] is None or isinstance(cfg['embedding_property'], List), "embedding_property must be a List or None"
    assert cfg['conservation'] is None or isinstance(cfg['conservation'], Dict), "cfg.model.config.conservations must be a Dict or None"
    assert cfg['universal_embedding'] is None or isinstance(cfg['universal_embedding'], Dict), "cfg.model.config.universal_embedding must be a Dict or None"

    # statistics
    if not isinstance(cfg['statistics'], List): 
        cfg['statistics'] = [cfg['statistics']]

    # Lmax, lmax
    if isinstance(cfg['Lmax'], int):
        cfg['Lmax'] = [cfg['Lmax']] * cfg['num_layers']
    if isinstance(cfg['lmax'], int):
        cfg['lmax'] = [cfg['lmax']] * cfg['num_layers']

    # radial_mlp
    if isinstance(cfg['radial_mlp']['hidden'][0], list): # not safe
        pass
    else:
        cfg['radial_mlp']['hidden'] = [cfg['radial_mlp']['hidden']] * cfg['num_layers']

    # readout_emlp
    if 'readout_mlp' in cfg['kwargs']:
        cfg['readout_emlp'] = cfg['kwargs']['readout_mlp']

    # embedding property
    cfg['embedding_property'] = cfg['embedding_property'] or []

    # atomic_numbers
    cfg['atomic_numbers'] = sorted(cfg['statistics'][0]['atomic_numbers'])

    # avg_num_neighbors
    cfg['avg_num_neighbors'] = cfg['statistics'][0]['avg_num_neighbors']

    if "energy" in cfg['target_property']: 
        cfg['atomic_energies'] = [stats['atomic_energy'] for stats in cfg['statistics']]
    else:
        cfg['atomic_energies'] = None

    # short_range and long_range
    cfg['short_range'] = cfg.get('short_range', {}) or SHORT_RANGE
    cfg['long_range'] = cfg.get('long_range', {}) or LONG_RANGE
 
    # inter
    if isinstance(cfg['inter']['l1l2'], str) or cfg['inter']['l1l2'] == None:
        cfg['inter']['l1l2'] = [cfg['inter']['l1l2']] * cfg['num_layers']

    # prod
    if isinstance(cfg['prod']['l1l2'], str) or cfg['prod']['l1l2'] == None:
        cfg['prod']['l1l2'] = [cfg['prod']['l1l2']] * cfg['num_layers']
    if isinstance(cfg['prod']['l3l1'], str) or cfg['prod']['l3l1'] == None:
        cfg['prod']['l3l1'] = [cfg['prod']['l3l1']] * cfg['num_layers']
    if isinstance(cfg['prod']['correlation'], int):
        cfg['prod']['correlation'] = [cfg['prod']['correlation']] * cfg['num_layers']


    # === uie and uee ===
    universal_embedding = cfg.get('universal_embedding', {})
    universal_embedding['invariant'] = universal_embedding.get('invariant', {})
    universal_embedding['equivariant'] = universal_embedding.get('equivariant', {})
    universal_embedding['invariant_embedding_property'] = []
    for k, v in universal_embedding['invariant'].items():
        if v.get('enable', False):
            universal_embedding['invariant_embedding_property'].append(k)
    universal_embedding['equivariant_embedding_property'] = []
    for k, v in universal_embedding['equivariant'].items():
        if v.get('enable', False):
            universal_embedding['equivariant_embedding_property'].append(k)

    # === conservation ===
    cfg['conservation'] = cfg.get('conservation', {}) or CONSERVATION

    return cfg


################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from __future__ import annotations
import logging
from typing import List, Dict
from pathlib import Path
import multiprocessing
from concurrent.futures import as_completed
from concurrent.futures import ProcessPoolExecutor


from ase import Atoms
from ase.io import read
from ase.db import connect
from ase.calculators.calculator import all_properties


from .split import random_split
from .quantity import KeySpecification


class DatasetsSplit:
    def __init__(self, train, valid, test):
        self.train = train
        self.valid = valid
        self.test = test

class ThreeDataset:
    def __init__(self, train, valid, test=None):
        self.data = DatasetsSplit(train=train, valid=valid, test=test or [])

    def __getitem__(self, idx):
        return [self.data.train, self.data.valid, self.data.test][idx]

    def __setitem__(self, idx, value):
        if idx == 0:
            self.data.train = value
        elif idx == 1:
            self.data.valid = value
        elif idx == 2:
            self.data.test = value
        else:
            raise IndexError("Index out of range for ThreeDataset")

    def __len__(self):
        return 3

    @property
    def train(self):
        return self.data.train

    @property
    def valid(self):
        return self.data.valid

    @property
    def test(self):
        return self.data.test


def check_keys(
    atomsList: List[Atoms],
    target_property: List[str],
    keyspec: KeySpecification,
    embedding_property: List[str] = [],
    check: bool = True,
):
    for atoms in atomsList:
        if check:
            if atoms.calc is not None:
                for p in (target_property + embedding_property):
                    found = False
                    if p in keyspec.info_keys:
                        key = keyspec.info_keys[p]
                        debug_key = f"{key}s"
                        if key in all_properties:
                            atoms.info[key] = atoms.calc.results[key]
                            found = True
                        if not found:
                            if debug_key in all_properties:
                                atoms.info[key] = atoms.calc.results[debug_key]
                    if p in keyspec.arrays_keys:
                        key = keyspec.arrays_keys[p]
                        debug_key = f"{key}s"
                        if key in all_properties:
                            atoms.arrays[key] = atoms.calc.results[key]
                            found = True
                        if not found:
                            if debug_key in all_properties:
                                atoms.arrays[key] = atoms.calc.results[debug_key]
    return atomsList


def read_file(fpath: str, target_property, keyspec, embedding_property):
    from ase.io import read
    from pathlib import Path

    fpath = Path(fpath)
    try:
        return check_keys(read(fpath, index=":"), target_property, keyspec, embedding_property)
    except Exception as e:
        logging.warning(f"Failed to read {fpath}: {e}, pass")
        return []

def ase_io_read(
    filename: str,
    target_property: List[str],
    keyspec,
    embedding_property: List[str],
    num_workers: int = None,
):
    """
    Behavior
    --------
    - If 'filename' is a file: read it normally using ASE.
    - If 'filename' is a directory: search for all '.xyz' and '.extxyz' files
      (recursively), read them in parallel (using 1/4 of available CPUs),
      and aggregate all structures.

    Parameters
    ----------
    filename : str
        Path to a file or directory.
    target_property : List[str]
        List of target properties to check.
    keyspec :
        Specification of property keys (type depends on your check_keys implementation).
    embedding_property : List[str]
        List of embedding-related properties.

    Returns
    -------
    list
        Aggregated structures passed to check_keys().
    """

    path = Path(filename)

    if num_workers is None:
        num_workers = max(1, multiprocessing.cpu_count() // 4)

    all_structures = []

    if path.is_dir():
        logging.info(f"'{path}' is a directory. Searching for all '.xyz' and '.extxyz' files recursively...")

        xyz_files = list(path.rglob("*.xyz")) + list(path.rglob("*.extxyz"))
        if not xyz_files:
            raise FileNotFoundError(f"No '.xyz' or '.extxyz' files found in directory '{path}'.")

        logging.info(f"Found {len(xyz_files)} files. Reading them in parallel...")
        logging.info(f"Using {num_workers} processes for parallel reading.")

        with ProcessPoolExecutor(max_workers=num_workers) as executor:
            futures = {
                executor.submit(read_file, str(f), target_property, keyspec, embedding_property): f
                for f in xyz_files
            }

            # for future in tqdm(as_completed(futures), total=len(futures), desc="Reading files"):
            #     all_structures.extend(future.result())

            for future in as_completed(futures):
                all_structures.extend(future.result())

        logging.info(f"Successfully read {len(all_structures)} structures from {len(xyz_files)} files.")
    else:
        all_structures = check_keys(read(path, index=":"), target_property, keyspec, embedding_property)

    return all_structures
    

def ase_db_connect(
    filename: str,
    target_property: List[str],
    keyspec: KeyboardInterrupt,
    embedding_property: List[str],
    num_workers: int = max(1, multiprocessing.cpu_count() // 4),
):
    return check_keys(
        [row.toatoms() for row in connect(filename).select()],
        target_property,
        keyspec,
        embedding_property,
    )


def fair_lmdb(filename: str):
    raise NotImplementedError("fair_lmdb is not yet implemented")

def torchsim_h5(filename: str):
    raise NotImplementedError("torchsim_h5 is not yet implemented")

HOW_TO_READ = {"ase": ase_io_read, "ase_db": ase_db_connect, "fair": fair_lmdb, "torchsim": torchsim_h5}


def _read(
    cfg: Dict,
    target_property: List[str],
    embedding_property: List[str],
    keyspec: KeyboardInterrupt,
    in_datamodule: bool = False,
) -> ThreeDataset:
    file_type = cfg.get("dataset", {}).get("type", "ase")
    no_valid_set = cfg.get("dataset", {}).get("no_valid_set", False)
    num_workers = cfg.get("dataset", {}).get("num_workers", max(1, multiprocessing.cpu_count() // 4))
    try:
        train_file = cfg["dataset"]["train_file"]
    except Exception as e:
        raise RuntimeError(
            "Failed to retrieve 'cfg.dataset.train_file' from configuration"
        ) from e
    valid_file = cfg.get("dataset", {}).get("valid_file", None)
    test_files = cfg.get("dataset", {}).get("test_files", None)
    try:
        tmp_train_atoms_list = HOW_TO_READ[file_type](
            train_file, target_property, keyspec, embedding_property, num_workers
        )
    except Exception as e:
        raise RuntimeError(
            f"Failed to load training file from cfg.dataset.train_file: {e}"
        )
    try:
        tmp_valid_atoms_list = (
            HOW_TO_READ[file_type](
                valid_file, target_property, keyspec, embedding_property, num_workers
            )
            if valid_file is not None
            else None
        )
    except Exception as e:
        raise RuntimeError(
            f"Failed to load validation file from cfg.dataset.valid_file: {e}"
        )
    try:
        if test_files is not None:
            if isinstance(test_files, str):
                test_atoms_list = [
                    HOW_TO_READ[file_type](
                        test_files, target_property, keyspec, embedding_property, num_workers
                    )
                ]
            elif isinstance(test_files, list):
                test_atoms_list = [
                    HOW_TO_READ[file_type](
                        f, target_property, keyspec, embedding_property, num_workers
                    )
                    for f in test_files
                ]
            else:
                test_atoms_list = None
        else:
            test_atoms_list = None
    except Exception as e:
        raise RuntimeError(f"Failed to load test file from cfg.dataset.test_files: {e}")
    
    if not in_datamodule:
        if test_atoms_list is None:
            logging.info("No test file is given")

    if tmp_valid_atoms_list is not None:
        train_atoms_list = tmp_train_atoms_list
        valid_atoms_list = tmp_valid_atoms_list
        if not in_datamodule:
            logging.info(
                f"Using training set from: {train_file}",
            )
            logging.info(
                f"Using validation set from: {valid_file}",
            )
            if test_atoms_list is not None:
                logging.info(
                    f"Using test set from: {test_files}",
                )
    elif cfg.get("dataset", {}).get("valid_from_index", False):
        # In the earlier version, the order of the training set was not taken 
        # into account, and therefore only the valid indices were saved.
        train_index_path = Path(".") / "train.index"
        valid_index_path = Path(".") / "valid.index"
        assert (
            valid_index_path.is_file()
        ), f"File does not exist or is not a regular file: {valid_index_path}"
        with valid_index_path.open("r", encoding="utf-8") as f:
            valid_indices = [int(line.strip()) for line in f if line.strip()]
        valid_atoms_list = [tmp_train_atoms_list[i] for i in valid_indices]
        if train_index_path.exists():
            assert (
                train_index_path.is_file()
            ), f"File does not exist or is not a regular file: {train_index_path}"
            with train_index_path.open("r", encoding="utf-8") as f:
                train_indices = [int(line.strip()) for line in f if line.strip()]
            train_atoms_list = [tmp_train_atoms_list[i] for i in train_indices]
        else:
            train_atoms_list = [
                item
                for idx, item in enumerate(tmp_train_atoms_list)
                if idx not in valid_indices
            ]
        if not in_datamodule:
            logging.info(
                f"Using training set from: {train_file}",
            )
            logging.info(f"Using valid set index from: {str(valid_index_path)}")
            if test_atoms_list is not None:
                logging.info(
                    f"Using test set from: {test_files}",
                )
    elif no_valid_set:
        train_atoms_list = tmp_train_atoms_list
        valid_atoms_list = None
        if not in_datamodule:
            logging.info(
                f"Using training set from: {train_file}",
            )
            logging.info(
                f"This training has no validation set, you must use lr_scheduler not depending on validation set",
            )
            if test_atoms_list is not None:
                logging.info(
                    f"Using test set from: {test_files}",
                )
    else:
        try:
            ratio = cfg["dataset"]["valid_ratio"]
        except Exception as e:
            raise RuntimeError(
                "Valid_ratio must be provided if no validation file is given."
            ) from e
        assert isinstance(
            ratio, float
        ), "Valid_ratio must be provided if no validation file is given"
        assert 0.0 < ratio < 1.0, "Valid_ratio must be in the range (0, 1)."
        if not in_datamodule:
            logging.info(
                f"Using training set from: {train_file}",
            )
            logging.info(
                "Using random %s%% of training set for validation",
                100 * ratio,
            )
            if test_atoms_list is not None:
                logging.info(
                    f"Using test set from: {test_files}",
                )
        train_atoms_list, valid_atoms_list = random_split(
            tmp_train_atoms_list,
            ratio,
            cfg.get("dataset", {}).get("split_seed", 1),
        )

    assert len(train_atoms_list) > 0, "Training set is empty !"
    if not no_valid_set:
        assert len(valid_atoms_list) > 0, "Validation set is empty !"

    return ThreeDataset(train_atoms_list, valid_atoms_list, test_atoms_list)



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
'''
from torch_ema
It's not a completely official implementation. We made a few modifications.
https://github.com/fadel/pytorch_ema
'''

'''
Copyright (c) 2019 Samuel G. Fadel <samuelfadel@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
'''

import weakref
import copy
import contextlib
from typing import Iterable, Optional

import torch


class ExponentialMovingAverage:
    def __init__(
        self,
        parameters: Iterable[torch.nn.Parameter],
        decay: float,
        use_num_updates: bool = True,
    ):
        if decay < 0.0 or decay > 1.0:
            raise ValueError("Decay must be between 0 and 1")
        self.decay = decay
        self.num_updates = 0 if use_num_updates else None
        parameters = list(parameters)
        self.shadow_params = [p.clone().detach() for p in parameters]
        self.collected_params = None
        # By maintaining only a weakref to each parameter,
        # we maintain the old GC behaviour of ExponentialMovingAverage:
        # if the model goes out of scope but the ExponentialMovingAverage
        # is kept, no references to the model or its parameters will be
        # maintained, and the model will be cleaned up.
        self._params_refs = [weakref.ref(p) for p in parameters]

    def _get_parameters(
        self, parameters: Optional[Iterable[torch.nn.Parameter]]
    ) -> Iterable[torch.nn.Parameter]:
        if parameters is None:
            parameters = [p() for p in self._params_refs]
            if any(p is None for p in parameters):
                raise ValueError(
                    "(One of) the parameters with which this "
                    "ExponentialMovingAverage "
                    "was initialized no longer exists (was garbage collected);"
                    " please either provide `parameters` explicitly or keep "
                    "the model to which they belong from being garbage "
                    "collected."
                )
            return parameters
        else:
            parameters = list(parameters)
            if len(parameters) != len(self.shadow_params):
                raise ValueError(
                    "Number of parameters passed as argument is different "
                    "from number of shadow parameters maintained by this "
                    "ExponentialMovingAverage"
                )
            return parameters

    def update(self, parameters: Optional[Iterable[torch.nn.Parameter]] = None) -> None:
        """
        Update currently maintained parameters.

        Call this every time the parameters are updated, such as the result of
        the `optimizer.step()` call.

        Args:
            parameters: Iterable of `torch.nn.Parameter`; usually the same set of
                parameters used to initialize this object. If `None`, the
                parameters with which this `ExponentialMovingAverage` was
                initialized will be used.
        """
        parameters = self._get_parameters(parameters)
        decay = self.decay
        if self.num_updates is not None:
            self.num_updates += 1
            decay = min(decay, (1 + self.num_updates) / (10 + self.num_updates))
        one_minus_decay = 1.0 - decay
        with torch.no_grad():
            for s_param, param in zip(self.shadow_params, parameters):
                tmp = s_param - param
                # tmp will be a new tensor so we can do in-place
                tmp.mul_(one_minus_decay)
                s_param.sub_(tmp)

    def copy_to(
        self, parameters: Optional[Iterable[torch.nn.Parameter]] = None
    ) -> None:
        """
        Copy current averaged parameters into given collection of parameters.

        Args:
            parameters: Iterable of `torch.nn.Parameter`; the parameters to be
                updated with the stored moving averages. If `None`, the
                parameters with which this `ExponentialMovingAverage` was
                initialized will be used.
        """
        parameters = self._get_parameters(parameters)
        for s_param, param in zip(self.shadow_params, parameters):
            param.data.copy_(s_param.data)

    def store(self, parameters: Optional[Iterable[torch.nn.Parameter]] = None) -> None:
        """
        Save the current parameters for restoring later.

        Args:
            parameters: Iterable of `torch.nn.Parameter`; the parameters to be
                temporarily stored. If `None`, the parameters of with which this
                `ExponentialMovingAverage` was initialized will be used.
        """
        parameters = self._get_parameters(parameters)
        self.collected_params = [param.clone() for param in parameters]

    def restore(
        self, parameters: Optional[Iterable[torch.nn.Parameter]] = None
    ) -> None:
        """
        Restore the parameters stored with the `store` method.
        Useful to validate the model with EMA parameters without affecting the
        original optimization process. Store the parameters before the
        `copy_to` method. After validation (or model saving), use this to
        restore the former parameters.

        Args:
            parameters: Iterable of `torch.nn.Parameter`; the parameters to be
                updated with the stored parameters. If `None`, the
                parameters with which this `ExponentialMovingAverage` was
                initialized will be used.
        """
        if self.collected_params is None:
            raise RuntimeError(
                "This ExponentialMovingAverage has no `store()`ed weights "
                "to `restore()`"
            )
        parameters = self._get_parameters(parameters)
        for c_param, param in zip(self.collected_params, parameters):
            param.data.copy_(c_param.data)

    @contextlib.contextmanager
    def average_parameters(
        self, parameters: Optional[Iterable[torch.nn.Parameter]] = None
    ):
        r"""
        Context manager for validation/inference with averaged parameters.

        Equivalent to:

            ema.store()
            ema.copy_to()
            try:
                ...
            finally:
                ema.restore()

        Args:
            parameters: Iterable of `torch.nn.Parameter`; the parameters to be
                updated with the stored parameters. If `None`, the
                parameters with which this `ExponentialMovingAverage` was
                initialized will be used.
        """
        parameters = self._get_parameters(parameters)
        self.store(parameters)
        self.copy_to(parameters)
        try:
            yield
        finally:
            self.restore(parameters)

    def to(self, device=None, dtype=None) -> None:
        r"""Move internal buffers of the ExponentialMovingAverage to `device`.

        Args:
            device: like `device` argument to `torch.Tensor.to`
        """
        # .to() on the tensors handles None correctly
        self.shadow_params = [
            (
                p.to(device=device, dtype=dtype)
                if p.is_floating_point()
                else p.to(device=device)
            )
            for p in self.shadow_params
        ]
        if self.collected_params is not None:
            self.collected_params = [
                (
                    p.to(device=device, dtype=dtype)
                    if p.is_floating_point()
                    else p.to(device=device)
                )
                for p in self.collected_params
            ]
        return

    def state_dict(self) -> dict:
        r"""Returns the state of the ExponentialMovingAverage as a dict."""
        # Following PyTorch conventions, references to tensors are returned:
        # "returns a reference to the state and not its copy!" -
        # https://pytorch.org/tutorials/beginner/saving_loading_models.html#what-is-a-state-dict
        return {
            "decay": self.decay,
            "num_updates": self.num_updates,
            "shadow_params": self.shadow_params,
            "collected_params": self.collected_params,
        }

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Loads the ExponentialMovingAverage state.

        Args:
            state_dict (dict): EMA state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        # deepcopy, to be consistent with module API
        state_dict = copy.deepcopy(state_dict)
        self.decay = state_dict["decay"]
        if self.decay < 0.0 or self.decay > 1.0:
            raise ValueError("Decay must be between 0 and 1")
        self.num_updates = state_dict["num_updates"]
        assert self.num_updates is None or isinstance(
            self.num_updates, int
        ), "Invalid num_updates"

        self.shadow_params = state_dict["shadow_params"]
        assert isinstance(self.shadow_params, list), "shadow_params must be a list"
        assert all(
            isinstance(p, torch.Tensor) for p in self.shadow_params
        ), "shadow_params must all be Tensors"

        self.collected_params = state_dict["collected_params"]
        if self.collected_params is not None:
            assert isinstance(
                self.collected_params, list
            ), "collected_params must be a list"
            assert all(
                isinstance(p, torch.Tensor) for p in self.collected_params
            ), "collected_params must all be Tensors"
            assert len(self.collected_params) == len(
                self.shadow_params
            ), "collected_params and shadow_params had different lengths"

        if len(self.shadow_params) == len(self._params_refs):
            # Consistant with torch.optim.Optimizer, cast things to consistant
            # device and dtype with the parameters
            params = [p() for p in self._params_refs]
            # If parameters have been garbage collected, just load the state
            # we were given without change.
            if not any(p is None for p in params):
                # ^ parameter references are still good
                for i, p in enumerate(params):
                    self.shadow_params[i] = self.shadow_params[i].to(
                        device=p.device, dtype=p.dtype
                    )
                    if self.collected_params is not None:
                        self.collected_params[i] = self.collected_params[i].to(
                            device=p.device, dtype=p.dtype
                        )
        else:
            raise ValueError(
                "Tried to `load_state_dict()` with the wrong number of "
                "parameters in the saved state."
            )



###############################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################


from typing import Dict, List

import torch


from .ictd import ICTD
from .utils import add_dict_to_left
from .paths import satisfy, generate_prod_paths
from .linear import SelfInteraction, LinearDict
from .einsum import ProdEinsumTC

PATH = 4
BATCH = 5
CHANNEL = 6


class SelfContraction(torch.nn.Module):
    def __init__(
        self,
        num_channel: int = 64,
        num_channel_hidden: int = 64,
        lmax_in: int = 3,
        ls_out: List[int] = 2,
        atomic_numbers: List[int] = [],
        prod: Dict = {},
        bias: bool = False,
        layer: int = -1,
        num_layers: int = 2,
    ) -> None:
        super().__init__()

        # === ICT ===
        for r in range(lmax_in + 1):
            DS = ICTD(r, r)[1]
            self.register_buffer(f"D_{r}_{r}_1", DS[0].to(torch.get_default_dtype()))
            del DS
        
        # === used in init === 
        
        l1l2 = prod.get("l1l2", [None] * num_layers)[layer]
        l3l1 = prod.get("l3l1", [None] * num_layers)[layer]
        correlation = prod.get("correlation", [3,] * num_layers)[layer]

        if isinstance(prod.get("element_aware", True), bool):
            element_aware = {}
            for l in ls_out:
                element_aware[l] = prod.get("element_aware", True)
        else:
             element_aware = prod["element_aware"]
        if isinstance(prod.get("coupled_channel", True), bool):
            coupled_channel = {}
            for l in ls_out:
                coupled_channel[l] = prod.get("coupled_channel", True)
        else:
            coupled_channel = prod["coupled_channel"]

        linear_type = {}
        for l in ls_out:
            linear_type[l] = (element_aware[l], not coupled_channel[l])

        # === init ===
        self.correlation = correlation
        self.lmax_in = lmax_in
        self.ls_out = ls_out
        self.layer = layer
        self.coupled_channel = coupled_channel

        # Below TODO for xzm
  
        # === prod ===
        self.paths_list_list, self.exprs_list_list = generate_prod_paths(
            lmax_in, ls_out, self.correlation, l1l2, None, l3l1,
        )
        
        filter_idx = prod.get("filter_idx", None)
        if filter_idx:
            assert self.correlation == 2, (
                "filter_idx can only be used when correlation = 2, "
                "because implementing higher-order cases is cumbersome"
            )
            filter_nu2_combs = []
            filter_nu2_exprs = []
            nu2_combs = self.paths_list_list[0]
            nu2_exprs = self.exprs_list_list[0]
            for idx, (comb, expr) in enumerate(zip(nu2_combs, nu2_exprs)):
                if idx not in filter_idx:
                    filter_nu2_combs.append(comb)
                    filter_nu2_exprs.append(expr)
            self.paths_list_list[0] = filter_nu2_combs
            self.exprs_list_list[0] = filter_nu2_exprs

        self.ctrs = torch.nn.ModuleList()
        for v in range(self.correlation - 1):
            ctrs = torch.nn.ModuleList()
            for comb, expr in zip(
                self.paths_list_list[v], self.exprs_list_list[v]
            ):
                ctrs.append(ProdEinsumTC((comb)))
            self.ctrs.append(ctrs)

        # === l3_count for each nu ===
        # nu = 1 
        nu_l3_count = {1: {l3: 0 for l3 in range(lmax_in + 1)}}  
        for l3 in range(lmax_in + 1):
            nu_l3_count[1][l3] += 1

        # nu > 1
        for nu in range(2, self.correlation + 1):
            nu_l3_count[nu] = {l3: 0 for l3 in range(lmax_in + 1)} 
            for l1 in range(lmax_in + 1):
                for l2 in range(lmax_in+ 1):
                    for l3 in range(abs(l1 - l2), min(lmax_in, l1 + l2) + 1, 2):
                        k = (l1 + l2 - l3) // 2
                        if satisfy(l1, l2, l1l2) and satisfy(l3, l1, l3l1):
                            if filter_idx:
                                for comb in filter_nu2_combs:
                                    this_comb = (l1, l2, l3 ,k)
                                    if this_comb == comb:
                                        nu_l3_count[nu][l3] += nu_l3_count[nu-1][l1]
                            else:
                                nu_l3_count[nu][l3] += nu_l3_count[nu-1][l1]


        # === lienar ===
        self.linears = torch.nn.ModuleDict()
        for nu in range(1, self.correlation+1):
            inner_dict = torch.nn.ModuleDict()
            for l3 in ls_out:
                if sum([nu_l3_count[nu][l3]]) > 0: 
                    linear_layer = LinearDict[linear_type[l3]](
                        num_channel_hidden * sum([nu_l3_count[nu][l3]]),
                        num_channel_hidden,
                        bias=(l3 == 0 and bias),
                        l=l3,
                        atomic_numbers=atomic_numbers,
                        groups=prod.get('groups', None),
                    )
                    inner_dict[str(l3)] = linear_layer
            self.linears[str(nu)] = inner_dict

        self.linear = SelfInteraction(
            in_channel=num_channel_hidden,
            out_channel=num_channel,
            ls=ls_out,
            bias=bias and layer == num_layers -1,
        )


    def D(self, l: int):
            return dict(self.named_buffers())[f"D_{l}_{l}_1"]
    
    def forward(
        self,
        node_feats: Dict[int, torch.Tensor],
        node_attrs: torch.Tensor,
        sc: Dict[int, torch.Tensor],
    ) -> Dict[int, torch.Tensor]:
        
        corr_feats = {
            0: {
                l: [node_feats[l]] for l in node_feats
            }
        }

        for nu, ctrs in enumerate(self.ctrs):
            corr_feats[nu + 1] = {l: [] for l in range(self.lmax_in + 1)}
            for idx, ctr in enumerate(ctrs):
                l1, l2, l3, _ = self.paths_list_list[nu][idx]
                tmp = ctr(
                    torch.stack(corr_feats[nu][l1], dim=0),
                    node_feats[l2],
                )

                P = tmp.size(0); B = tmp.size(1); C = tmp.size(2)

                tmp = torch.bmm(
                    tmp.reshape(P, B * C, -1), self.D(l3).repeat(P, 1, 1)
                ).reshape((P, B, C) + (3,) * l3)

                tmp = torch.unbind(tmp, dim=0)
                corr_feats[nu + 1][l3].extend(tmp)

        out_dict = {}
        for nu_str, linears in self.linears.items():
            nu = int(nu_str)

            for l3_str, linear in linears.items():
                l3 = int(l3_str)

                if self.coupled_channel[l3]:
                    merged = torch.cat([t for t in corr_feats[nu-1][l3]], dim=1)
                else:
                    merged = torch.stack([t for t in corr_feats[nu-1][l3]], dim=0)

                out = linear(merged, node_attrs)

                if l3 in out_dict:
                    out_dict[l3] += out
                else:
                    out_dict[l3] = out

        return add_dict_to_left(self.linear(out_dict), sc)



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import Dict, Tuple, Optional, NamedTuple


import torch


from ...utils.torch_scatter import scatter_sum


def expand_dims_to(T: torch.Tensor, n_dim: int, dim: int = -1) -> torch.Tensor:
    '''jit-safe'''
    while T.ndim < n_dim:
        T = T.unsqueeze(dim)
    return T


def add_dict_to_left(
    T1: Dict[int, torch.Tensor], T2: Dict[int, torch.Tensor]
) -> Dict[int, torch.Tensor]:

    for k in T2:
        if k in T1:
            T1[k] = T1[k] + T2[k]
        else:
            T1[k] = T2[k]
    return T1


def add_dict_to_right(
        T1: Dict[int, torch.Tensor], T2: Dict[int, torch.Tensor]
    ) -> Dict[int, torch.Tensor]:

    for k in T1:
        if k in T2:
            T2[k] = T2[k] + T1[k]
        else:
            T2[k] = T1[k]
    return T2


def compute_fixed_charge_dipole(
    charges: torch.Tensor,
    positions: torch.Tensor,
    batch: torch.Tensor,
    num_graphs: int,
) -> torch.Tensor:
    mu = positions * charges.unsqueeze(-1) * 4.8032047  # e·Å to Debye
    return scatter_sum(src=mu, index=batch.unsqueeze(-1), dim=0, dim_size=num_graphs)


def torch_full_3x3_to_voigt_6_stress(stress_tensor: torch.Tensor) -> torch.Tensor:
    """
    Convert stress tensor [batch, 3, 3] -> [batch, 6] in Voigt notation,
    matching ASE's full_3x3_to_voigt_6_stress.
    """
    s = stress_tensor
    s_voigt = torch.stack(
        [
            s[..., 0, 0],  # σ_xx
            s[..., 1, 1],  # σ_yy
            s[..., 2, 2],  # σ_zz
            0.5 * (s[..., 1, 2] + s[..., 2, 1]),  # σ_yz
            0.5 * (s[..., 0, 2] + s[..., 2, 0]),  # σ_xz
            0.5 * (s[..., 0, 1] + s[..., 1, 0]),  # σ_xy
        ],
        dim=-1,
    )
    return s_voigt

def select_corresponding_level(
        x: torch.Tensor, node_level: torch.Tensor, num_levels: int
    ) -> torch.Tensor:
    B = x.size(0)
    C_LEVELS = x.size(1)
    mask = torch.zeros(B, num_levels, C_LEVELS // num_levels, device=x.device, dtype=x.dtype)
    idx = torch.arange(B, device=x.device, dtype=torch.int64)
    mask[idx, node_level, :] = 1
    mask = mask.reshape((B, C_LEVELS))
    # return x * mask
    return x * expand_dims_to(mask, x.ndim, -1)


class Graph(NamedTuple):
    lmp: bool
    lmp_data: Optional[torch.Tensor]
    lmp_natoms: Tuple[int, int]
    num_graphs: int
    displacement: Optional[torch.Tensor]
    positions: torch.Tensor
    edge_vector: torch.Tensor
    edge_length: torch.Tensor
    lattice: torch.Tensor
    node_level: torch.Tensor
    num_atoms_arange: torch.Tensor


class LAMMPS_MP(torch.autograd.Function):
    @staticmethod
    def forward(ctx, *args):
        feats, data = args  # unpack
        ctx.vec_len = feats.shape[-1]
        ctx.data = data
        out = torch.empty_like(feats)
        data.forward_exchange(feats, out, ctx.vec_len)
        return out

    @staticmethod
    def backward(ctx, *grad_outputs):
        (grad,) = grad_outputs  # unpack
        gout = torch.empty_like(grad)
        ctx.data.reverse_exchange(grad, gout, ctx.vec_len)
        return gout, None
    

def dict2flatten(max_r: int, t: Dict[int, torch.Tensor]):
    tmp = []
    B, C = t[0].shape[:2]
    for k in sorted(t.keys()):
        flat = t[k].reshape(B, C, -1) 
        tmp.append(flat)
    return torch.cat(tmp, dim=-1).reshape(B, -1)


def flatten2dict(max_r: int, t: torch.Tensor, C: int) -> Dict[int, torch.Tensor]:
    B = t.size(0)
    ndim = (3 ** (max_r + 1) - 1) // 2
    t = t.reshape(B, C, ndim)  
    outs = {}
    start_idx = 0
    for r in range(max_r+1):
        shape = (B, C,) + (3,) * r
        delta = 3**r
        outs[r] = t[:, :, start_idx:start_idx+delta].reshape(shape)
        start_idx += delta
    return outs





################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import logging
from enum import Enum
from typing import Dict, List, Optional, Tuple
from dataclasses import dataclass, field


from .utils import (
    voigt_to_matrix,
    shape_fn_for_hessians,
    default_value_for_hessians,
    default_value_for_rank0_atom,
    default_value_for_rank1_atom,
    default_value_for_rank2_atom,
    default_value_for_rank3_atom,
    default_value_for_rank4_atom,
    default_value_for_rank0_graph,
    default_value_for_rank1_graph,
    default_value_for_rank2_graph,
    default_value_for_rank3_graph,
    default_value_for_rank4_graph,
)


PROPERTY = {
    "level": {
        "ase_name": None,
        'type': 'int',
        "scope": "per-system",
        "rank": 0,
        "abbreviation": "LEVEL",
        "shape": {
            "in_data": (1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "energy": {
        "ase_name": "energy",
        'type': 'float',
        "scope": "per-system",
        "rank": 0,
        "abbreviation": "E",
        "shape": {
            "in_data": (1,),
            "shape_fn": None,
        },

        "default_value_fn": default_value_for_rank0_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "forces": {
        "ase_name": "forces",
        'type': 'float',
        "scope": "per-atom",
        "rank": 1,
        "abbreviation": "F",
        "shape": {
            "in_data": (-1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_atom,
        "must_be_with": {
            1: ['energy'],
        },
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False, 
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['positions'],
    },
    "edge_forces": {
        "ase_name":None,
        'type': 'float',
        "scope": "per-edge",
        "rank": 1,
        "abbreviation": "EDGE_F",
        "shape": {
            "in_data": (-1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_atom, # placeholder
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": False, # can be embedded through uee to achice DeNs
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['edge_vector'],
    },
    "direct_forces": {
        "ase_name": "forces",
        'type': 'float',
        "scope": "per-atom",
        "rank": 1,
        "abbreviation": "D_F",
        "shape": {
            "in_data": (-1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_atom,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False, 
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "hessians": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-edge",
        "rank": 2,
        "abbreviation": "HESSIAN",
        "shape": {
            "in_data": (-1,),
            "shape_fn": shape_fn_for_hessians,
        },
        "default_value_fn": default_value_for_hessians,
        "must_be_with": {
            1: ['energy', 'forces'],
        },
        "conflict_with": {},
        "enable_prediction": True,
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": True,
        "requires_grad_with": ['positions'],
    },
    "direct_hessians": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-edge",
        "rank": 2,
        "abbreviation": "D_HESSIAN",
        "shape": {
            "in_data": (-1,),
            "shape_fn": shape_fn_for_hessians,
        },
        "default_value_fn": default_value_for_hessians,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "stress": {
        "ase_name": "stress",
        'type': 'float',
        "scope": "per-system",
        "rank": 2,
        "abbreviation": "S",
        "shape": {
            "in_data": (1, 3, 3),
            "shape_fn": voigt_to_matrix,
        },
        "default_value_fn": default_value_for_rank2_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        # "requires_grad_with": ['displacement'],
        "requires_grad_with": [], # manual set
    },
    "direct_stress": {
        "ase_name": "stress",
        'type': 'float',
        "scope": "per-system",
        "rank": 2,
        "abbreviation": "D_S",
        "shape": {
            "in_data": (1, 3, 3),
            "shape_fn": voigt_to_matrix,
        },
        "default_value_fn": default_value_for_rank2_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "virials": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 2,
        "abbreviation": "V",
        "shape": {
            "in_data": (1, 3, 3),
            "shape_fn": voigt_to_matrix,
        },
        "default_value_fn": default_value_for_rank2_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        # "requires_grad_with": ['displacement'],
        "requires_grad_with": [], # manual set
    },
    "direct_virials": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 2,
        "abbreviation": "D_V",
        "shape": {
            "in_data": (1, 3, 3),
            "shape_fn": voigt_to_matrix,
        },
        "default_value_fn": default_value_for_rank2_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "atomic_stresses": {
        "ase_name": "stresses",
        'type': 'float',
        "scope": "per-atom",
        "rank": 2,
        "abbreviation": "A_S",
        "shape": {
            "in_data": (-1, 3, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank2_atom,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['edge_vector'],
    },
    "atomic_virials": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-atom",
        "rank": 2,
        "abbreviation": "A_V",
        "shape": {
            "in_data": (-1, 3, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank2_atom,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['edge_vector'],
    },
    "direct_dipole": {
        "ase_name": "dipole",
        'type': 'float',
        "scope": "per-system",
        "rank": 1,
        "abbreviation": "D",
        "shape": {
            "in_data": (1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_graph,
        "must_be_with": {
            1: ['charges'],
        },
        "conflict_with": {
            1: ["polarization"]
        },
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "conservative_dipole": {
        "ase_name": "dipole",
        'type': 'float',
        "scope": "per-system",
        "rank": 1,
        "abbreviation": "D",
        "shape": {
            "in_data": (1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_graph,
        "must_be_with": {},
        "conflict_with": {
            1: ["polarization"]
        },
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['electric_field'],
    },
    "polarization": {
        "ase_name": "polarization",
        'type': 'float',
        "scope": "per-atom",
        "rank": 1,
        "abbreviation": "P",
        "shape": {
            "in_data": (1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_graph,
        "must_be_with": [], 
        "must_be_with": {},
        "conflict_with": {
            1: ["direct_dipole", "conservative_dipole"]
        },
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['electric_field'],
    },
    "direct_polarizability": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 2,
        "abbreviation": "ALPHA",
        "shape": {
            "in_data": (1, 3, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank2_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "conservative_polarizability": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 2,
        "abbreviation": "ALPHA",
        "shape": {
            "in_data": (1, 3, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank2_graph,
        "must_be_with": {
            1: ["direct_dipole"],
            2: ["conservative_dipole"],
            3: ["polarization"]
        },
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": True,
        "requires_grad_with": ['electric_field'],
    },
    "born_effective_charges": {
        "ase_name": "born_effective_charges",
        'type': 'float',
        "scope": "per-atom",
        "rank": 2,
        "abbreviation": "BEC",
        "shape": {
            "in_data": (-1, 3, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank2_atom,
        "must_be_with": {
            1: ["direct_dipole"],
            2: ["conservative_dipole"],
            3: ["polarization"]
        },
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": True,
        "requires_grad_with": ['electric_field', 'positions'],
    },
    "magnetization": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 1,
        "abbreviation": "M",
        "shape": {
            "in_data": (1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['magnetic_field'],
    },
    "magnetic_susceptibility": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 2,
        "abbreviation": "CHI_M",
        "shape": {
            "in_data": (1, 3, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank2_graph,
        "must_be_with": {
            1: ['magnetization']
        },
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": True,
        "requires_grad_with": ['magnetic_field'],
    },
    "charges": {
        "ase_name": "charges",
        'type': 'float',
        "scope": "per-atom",
        "rank": 0,
        "abbreviation": "C",
        "shape": {
            "in_data": (-1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_atom,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "total_charge": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 0,
        "type": "graph",
        "abbreviation": "TC",
        "shape": {
            "in_data": (1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "spin_multiplicity": {
        "ase_name": None,
        'type': 'int',
        "scope": "per-system",
        "rank": 0,
        "abbreviation": "SM",
        "shape": {
            "in_data": (1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "initial_collinear_magmoms": {
        "ase_name": "initial_magmoms",
        'type': 'float',
        "scope": "per-atom",
        "rank": 0,
        "abbreviation": "I_C_MAG",
        "shape": {
            "in_data": (-1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_atom,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "initial_noncollinear_magmoms": {
        "ase_name": "initial_magmoms",
        'type': 'float',
        "scope": "per-atom",
        "rank": 1,
        "abbreviation": "I_NC_MAG",
        "shape": {
            "in_data": (-1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_atom,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "final_collinear_magmoms": {
        "ase_name": "magmoms",
        'type': 'float',
        "scope": "per-atom",
        "rank": 0,
        "abbreviation": "F_C_MAG",
        "shape": {
            "in_data": (-1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_atom,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "final_noncollinear_magmoms": {
        "ase_name": "magmoms",
        'type': 'float',
        "scope": "per-atom",
        "rank": 1,
        "abbreviation": "F_NC_MAG",
        "shape": {
            "in_data": (-1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_atom,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "collinear_magnetic_forces": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-atom",
        "rank": 0,
        "abbreviation": "C_MAG_F",
        "shape": {
            "in_data": (-1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_atom,
        "must_be_with": {
            1: ["magmoms_0"]
        },
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['magmoms_0'],
    },
    "noncollinear_magnetic_forces": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-atom",
        "rank": 1,
        "abbreviation": "NC_MAG_F",
        "shape": {
            "in_data": (-1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_atom,
        "must_be_with": {
            1: ["magmoms_1"]
        },
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": True,
        "second_derivative": False,
        "requires_grad_with": ['magmoms_1'],
    },
    "total_collinear_magmom": {
        "ase_name": "magmoms",
        'type': 'float',
        "scope": "per-system",
        "rank": 0,
        "abbreviation": "TCM",
        "shape": {
            "in_data": (1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "total_noncollinear_magmom": {
        "ase_name": "magmoms",
        'type': 'float',
        "scope": "per-system",
        "rank": 1,
        "abbreviation": "TNCM",
        "shape": {
            "in_data": (1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": True,   
        "enable_embedding": False,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "electric_field": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 1,
        "abbreviation": "EF",
        "shape": {
            "in_data": (1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "magnetic_field": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 1,
        "abbreviation": "MF",
        "shape": {
            "in_data": (1, 3),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank1_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "temperature": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 0,
        "abbreviation": "TEMP",
        "shape": {
            "in_data": (1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "electron_temperature": {
        "ase_name": None,
        'type': 'float',
        "scope": "per-system",
        "rank": 0,
        "abbreviation": "E_TEMP",
        "shape": {
            "in_data": (1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
    "spin_on": {
        "ase_name": None,
        'type': 'int',
        "scope": "per-system",
        "rank": 0,
        "abbreviation": "SPIN_ON",
        "shape": {
            "in_data": (1,),
            "shape_fn": None,
        },
        "default_value_fn": default_value_for_rank0_graph,
        "must_be_with": {},
        "conflict_with": {},
        "enable_prediction": False,   
        "enable_embedding": True,
        "first_derivative": False,
        "second_derivative": False,
        "requires_grad_with": [],
    },
}

SUPPORT_PREDICT_PROPERTY = [k for k, v in PROPERTY.items() if v['enable_prediction']]
UNIVERSAL_EMBEDDING_ALLOWED_PROPERTY = [k for k, v in PROPERTY.items() if v['enable_embedding']]

KEYS = {f"{k}_key": k for k in PROPERTY}

# should be delete in future versions TODO
class DefaultKeys(Enum):
    # basic
    ENERGY = "energy"
    FORCES = "forces"
    STRESS = "stress"
    VIRIALS = "virials"

    HESSIANS = "hessians"
    EDGE_FORCES = "edge_forces"
    ATOMIC_VIRIALS = "atomic_virials"
    ATOMIC_STRESSES = "atomic_stresses"
    
    # direct property
    DIRECT_FORCES = "direct_forces"
    DIRECT_STRESS = "direct_stress"
    DIRECT_VIRIALS = "direct_virials"
    DIRECT_DIPOLE = "direct_dipole"
    DIRECT_POLARIZABILITY = "direct_polarizability"
    DIRECT_HESSIANS = "direct_hessians"

    # charges
    CHARGES = "charges"
    TOTAL_CHARGE = "total_charge"

    # external field
    ELECTRIC_FIELD = "electric_field"
    MAGNETIC_FIELD = "magnetic_field"

    CONSERVATIVE_DIPOLE = "conservative_dipole"
    CONSERVATIVE_POLARIZABILITY = "conservative_polarizability"
    BORN_EFFECTIVE_CHARGES = "born_effective_charges" # do not consider LES
    MAGNETIZATION = "magnetization"
    MAGNETIC_SUSCEPTIBILITY = "magnetic_susceptibility"
    POLARIZATION = "polarization"

    # MAG
    INITIAL_COLLINEAR_MAGMOMS = "initial_collinear_magmoms"
    INITIAL_NONCOLLINEAR_MAGMOMS = "initial_noncollinear_magmoms"
    FINAL_COLLINEAR_MAGMOMS = "final_collinear_magmoms"
    FINAL_NONCOLLINEAR_MAGMOMS = "final_noncollinear_magmoms"
    COLLINEAR_MAGNETIC_FORCES = "collinear_magnetic_forces"
    NONCOLLINEAR_MAGNETIC_FORCES = "noncollinear_magnetic_forces"
    TOTAL_COLLINEAR_MAGMOM = "total_collinear_magmom"
    TOTAL_NONCOLLINEAR_MAGMOM = "total_noncollinear_magmom"
    SPIN_ON = "spin_on"

    # only for embedding
    LEVEL = 'level'
    TEMPERATURE = "temperature"
    ELECTRON_TEMPERATURE = "electron_temperature"
    SPIN_MULTIPLICITY = "spin_multiplicity"

    @staticmethod
    def keydict() -> dict[str, str]:
        key_dict = {}
        for member in DefaultKeys:
            key_name = f"{member.name.lower()}_key"
            key_dict[key_name] = member.value
        return key_dict


@dataclass
class KeySpecification:
    '''Modify from MACE to simplify reading property'''
    info_keys: Dict[str, str] = field(default_factory=dict)
    arrays_keys: Dict[str, str] = field(default_factory=dict)

    def update(
        self,
        info_keys: Optional[Dict[str, str]] = None,
        arrays_keys: Optional[Dict[str, str]] = None,
    ):
        if info_keys is not None:
            self.info_keys.update(info_keys)
        if arrays_keys is not None:
            self.arrays_keys.update(arrays_keys)
        return self

    @classmethod
    def from_defaults(cls):
        instance = cls()
        return update_keyspec_from_kwargs(instance, DefaultKeys.keydict())


def update_keyspec_from_kwargs(
    keyspec: KeySpecification, keydict: Dict[str, str]
) -> KeySpecification:
    '''Modify from MACE to simplify reading property'''
    infos = [f"{k}_key" for k, v in PROPERTY.items() if (v['scope'] != 'per-atom')]
    arrays = [f"{k}_key" for k, v in PROPERTY.items() if (v['scope'] == 'per-atom')]
    info_keys = {}
    arrays_keys = {}
    for key in infos:
        if key in keydict:
            info_keys[key[:-4]] = keydict[key]
    for key in arrays:
        if key in keydict:
            arrays_keys[key[:-4]] = keydict[key]
    keyspec.update(info_keys=info_keys, arrays_keys=arrays_keys)
    return keyspec


def get_target_property(cfg: Dict) -> List[str]:
    """
    Automatically infer the physical quantities required for training from the loss function.
    Ensure no conflicting physical quantities appear simultaneously.
    """
    try:
        loss = str(cfg["loss"]["_target_"]).lower()
    except KeyError as e:
        raise KeyError("Missing 'cfg.loss._target_' field in configuration") from e
    loss_property = cfg["loss"].get("loss_property", None)
    if loss_property is None:
        logging.warning(
            "The argument `cfg.loss.loss_property` must be provided by the user. "
            "It is kept optional only for backward compatibility with earlier versions. "
            "Omitting it may lead to bugs when predicted physical property names share common prefixes or substrings."
        )
        loss_property = []
        for p in SUPPORT_PREDICT_PROPERTY:
            clean_property = p.replace("_", "")
            clean_property = p
            if clean_property.lower() in loss:
                loss_property.append(p)

        for p in loss_property:
            conflict_with = PROPERTY[p]['conflict_with']
            for _, conflict_ps in conflict_with.items():
                for conflict_p in conflict_ps:
                    if conflict_p in loss_property:
                        raise ValueError(
                            f"Conflict Property detected: {p} with {conflict_with} cannot be used together."
                        )
    else:
        assert set(loss_property).issubset(list(PROPERTY))

    return loss_property


def get_embedding_property(cfg: Dict, separate: bool = False) -> List[str] | Tuple[List[str]]:
    invariant = cfg["model"]["config"].get("universal_embedding", {}).get("invariant", {})
    invariant_embedding_property = []
    for k, v in invariant.items():
        if v.get('enable', False):
            invariant_embedding_property.append(k)

    equivariant = cfg["model"]["config"].get("universal_embedding", {}).get("equivariant", {})
    equivariant_embedding_property = []
    for k, v in equivariant.items():
        if v.get('enable', False):
            equivariant_embedding_property.append(k)

    if separate:
        return invariant_embedding_property, equivariant_embedding_property
    else:
        return invariant_embedding_property + equivariant_embedding_property


# For Metrics
MAE_PROPERTY = [
        p for p in SUPPORT_PREDICT_PROPERTY 
        if p != "polarization" 
        and p != "final_collinear_magmoms"
        and p != "hessians"
    ]
RMSE_PROPERTY = [   
        p for p in SUPPORT_PREDICT_PROPERTY 
        if p != "polarization" 
        and p != "final_collinear_magmoms"
        and p != "hessians"
    ]
MAE_PER_ATOM_PROPERTY = [
    p for p, v in PROPERTY.items() 
    if v["scope"] == "per-system" 
    and p != "polarization"
    and p != "final_collinear_magmoms"
    and p != "stress"   
    and p != "direct_stress"  
        and p != "hessians"
]
RMSE_PER_ATOM_PROPERTY = [
    p for p, v in PROPERTY.items() 
    if v["scope"] == "per-system" 
    and p != "polarization" 
    and p != "final_collinear_magmoms"
    and p != "stress"    
    and p != "direct_stress"   
    and p != "hessians" 
]


fields = {f"compute_{k}": False for k, v in PROPERTY.items()}
@dataclass
class ComputeFlag:
    __annotations__ = {k: bool for k in fields} 
    locals().update(fields)


def get_target_irreps(target_property: List[str]):
    target_irreps = []
    target_irreps.extend([0]) if "energy" in target_property else None
    target_irreps.extend([0]) if "final_collinear_magmoms" in target_property else None
    target_irreps.extend([1]) if "final_noncollinear_magmoms" in target_property else None
    target_irreps.extend([0, 2]) if "direct_polarizability" in target_property else None
    target_irreps.extend([0, 2]) if "direct_stress" in target_property or \
        "direct_virials" in target_property else None
    target_irreps.extend([1]) if "direct_dipole" in target_property else None
    target_irreps.extend([1]) if "direct_forces" in target_property else None
    target_irreps.extend([0, 2]) if "direct_hessians" in target_property else None
    return sorted(list(set(target_irreps)))



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################


from typing import Optional, List, Tuple
from collections import defaultdict


from ...utils._global import LETTERS


def satisfy(l1: int, l2: int, restriction: Optional[str] = None) -> bool:
    if restriction == None:
        return True
    elif restriction == "<":
        return l1 < l2
    elif restriction == "<=":
        return l1 <= l2
    elif restriction == ">":
        return l1 > l2
    elif restriction == ">=":
        return l1 >= l2
    elif restriction == "==":
        return l1 == l2
    elif restriction == "!=":
        return l1 != l2
    else:
        raise ValueError(f"Unknown restriction: {restriction}")


def generate_combinations(
    l1max: int,
    l2max: int,
    l3max: int,
    l1l2: Optional[str] = None,
    l2l3: Optional[str] = None,
    l3l1: Optional[str] = None,
) -> List[Tuple]:
    combs = []
    for l1 in range(l1max + 1):
        for l2 in range(l2max + 1):
            for l3 in range(abs(l1 - l2), min(l3max, l1 + l2) + 1, 2):
                if (satisfy(l1, l2, l1l2) and satisfy(l2, l3, l2l3) and satisfy(l3, l1, l3l1)):
                    k = (l1 + l2 - l3) // 2
                    combs.append((l1, l2, l3, k))
    return combs


def generate_path(
    l1: int,
    l2: int,
    l3: int,
    k : int,
    add_batch_and_channel: bool = True,
) -> str:
    assert isinstance(l1, int) and l1 >= 0, "l1 must be a non-negative integer"
    assert isinstance(l2, int) and l2 >= 0, "l2 must be a non-negative integer"
    assert isinstance(l3, int) and l3 >= 0, "l3 must be a non-negative integer"
    assert isinstance(l3, int) and k >= 0, "k must be a non-negative integer"
    assert (l1 + l2 - l3) % 2 == 0, "Illegal contraction combination"

    in1_list = LETTERS[:l1]
    in2_list = LETTERS[l1 : l1 + l2]
    in1_list_copy = in1_list.copy()
    in2_list_copy = in2_list.copy()

    in1_idx = list(range(l1 - k, l1))
    in2_idx = list(range(k))
    
    for i in range(k):
        in2_list_copy[in2_idx[i]] = in1_list_copy[in1_idx[i]]

    in1 = "".join(in1_list_copy)
    in2 = "".join(in2_list_copy)
    out = "".join(
        [l for i, l in enumerate(in1_list_copy) if i not in in1_idx] 
        + 
        [l for _, l in enumerate(in2_list_copy) if l not in in1_list_copy]
    )

    if add_batch_and_channel:
        in1 = "bc" + in1
        in2 = "bc" + in2
        out = "bc" + out

    einsum_expr = f"{in1},{in2}->{out}"
    return einsum_expr


def generate_prod_paths(
    lmax_in, ls_out, correlation, l1l2, l2l3, l3l1,
):
    paths_list_list = []
    exprs_list_list = []
    current_ranks = set(range(lmax_in + 1))

    for nu in range(2, correlation + 1):
        path_list = [] # TODO convert this to dict
        expr_list = [] # TODO convert this to dict
        next_ranks = set()
        for l1 in current_ranks:
            for l2 in range(lmax_in + 1):
                lmin = abs(l1 - l2)
                lmax = min(lmax_in, l1 + l2)
                for l3 in range(lmin, lmax + 1, 2):
                    if satisfy(l1, l2, l1l2) and satisfy(l2, l3, l2l3) and satisfy(l3, l1, l3l1):
                        k = (l1 + l2 - l3) // 2
                        path_list.append((l1, l2, l3, k))
                        next_ranks.add(l3)
                        expr_list.append(generate_path(l1, l2, l3, k, True))
         
        paths_list_list.append(path_list)
        exprs_list_list.append(expr_list)
        current_ranks = next_ranks

    valid_ranks = ls_out
    for nu in reversed(range(correlation - 1)):
        filtered_path = []
        filtered_expr = []
        next_valid_ranks = set()

        for (l1, l2, l3, k), expr in zip(paths_list_list[nu], exprs_list_list[nu]):
            if l3 in valid_ranks:
                filtered_path.append((l1, l2, l3, k))
                filtered_expr.append(expr)
                next_valid_ranks.update([l1, l2])

        paths_list_list[nu] = filtered_path
        exprs_list_list[nu] = filtered_expr
        valid_ranks = next_valid_ranks

    return paths_list_list, exprs_list_list



def count_irreps(
        combs: List[Tuple],
        ictp_lower_weight: bool = False,
        ictc_lower_weight: bool = False,
        ictp_highest_weight: bool = True, 
        ictc_highest_weight: bool = True, 
    ):

    l3_count = defaultdict(int) 
    for comb in combs:
        l1, l2, _, k = comb 
        for l3 in range(abs(l1 - l2), l1 + l2 - 2 * k + 1):
            hw = (l3 == l1 + l2 - 2 * k)
            lw = (l3 <  l1 + l2 - 2 * k)
            if k == 0:
                if ictp_highest_weight and hw:
                    l3_count[l3] += 1
                if ictp_lower_weight and lw:
                    l3_count[l3] += 1
            else:
                if ictc_highest_weight and hw:
                    l3_count[l3] += 1
                if ictc_lower_weight and lw:
                    l3_count[l3] += 1
                    
    return dict(sorted(l3_count.items(), key=lambda x: x[0]))



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import packaging
from string import ascii_letters
from pathlib import Path

import torch

CACHE_DIR = Path.home() / ".cache" / "tace"
CACHE_DIR.mkdir(parents=True, exist_ok=True)


BOOL = {
    0: False,
    1: True,
}


DTYPE = {
    None: None,
    16: torch.float16,
    32: torch.float32,
    64: torch.float64,
    "fp16": torch.float16,
    "fp32": torch.float32,
    "fp64": torch.float64,
    "float16": torch.float16,
    "float32": torch.float32,
    "float64": torch.float64,
    torch.float16: torch.float16,
    torch.float32: torch.float32,
    torch.float64: torch.float64,
}


num_gpus = 128
DEVICE = {
    None: None,
    "cpu": torch.device("cpu"),
    "cuda": torch.device("cuda"),
    torch.device("cpu"): torch.device("cpu"),
    torch.device("cuda"): torch.device("cuda"),
    **{i: torch.device(f"cuda:{i}") for i in range(num_gpus)},
    **{f"cuda:{i}": torch.device(f"cuda:{i}") for i in range(num_gpus)},
    **{torch.device(f"cuda:{i}"): torch.device(f"cuda:{i}") for i in range(num_gpus)},
}

LETTERS = list(ascii_letters)[3:]


# _TORCH_VERSION = packaging.version.parse(torch.__version__)
# _TORCH_GE_2_9 = _TORCH_VERSION >= packaging.version.parse("2.9")
# _GLOBAL_STATE_INITIALIZED = False




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
# Description: This file contains linear readout function for tensor rank > 0
################################################################################

from math import sqrt
from typing import List, Optional


import torch


from .act import ACT
from .linear import Linear
from .mlp import LinearLayer
from .gate import NormGate
from .utils import select_corresponding_level


class ScalarReadOut(torch.nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        hidden_dim: List[int] = [],
        act: Optional[str] = "silu",
        bias: bool = False,
        forward_weight_init: bool = True,
        enable_layer_norm: bool = False,
        num_levels: int = 1,
        use_multi_head: bool = False,
    ):
        super().__init__()
        self.bias = bias
        self.dims = [in_dim] + hidden_dim + [out_dim]
        self.num_layers = len(self.dims) - 1
        assert self.num_layers >= 1
        act = ACT[act]()
        self.is_nonlinear = False
        self.enable_layer_norm = enable_layer_norm
        self.num_levels = num_levels
        self.use_multi_head = (use_multi_head) and (num_levels > 1) and len(hidden_dim) > 0
        if self.use_multi_head:
            assert len(hidden_dim) == 1, 'For multihead training, cfg.model.config.readout_mlp.hidden_dim must be only one neuron'

        # === build the MLP + weight init ===
        mlp = []
        for layer, (h_in, h_out) in enumerate(zip(self.dims, self.dims[1:])):

            # === weight initialization ===
            if forward_weight_init:
                norm_dim = h_in
                gain = 1.0 if act is None or (layer == 0) else sqrt(2)
            else:
                norm_dim = h_out
                gain = (
                    1.0 if act is None or (layer == self.num_layers - 1) else sqrt(2)
                )

            # === instantiate Linear ===
            linear_layer = LinearLayer(
                in_dim=h_in,
                out_dim=h_out,
                alpha=gain / sqrt(norm_dim),
                bias=bias,
            )
            mlp.append(linear_layer)

            # === optional LayerNorm ===
            if enable_layer_norm:
                if layer < len(self.dims) -2:
                    mlp.append(torch.nn.LayerNorm(h_out))
            del gain, norm_dim

            # === act ===
            if (layer != self.num_layers - 1) and (act is not None):
                mlp.append(act)
                self.is_nonlinear = True

        if self.use_multi_head: 
            self.mlp_1 = torch.nn.Sequential(*mlp[:-1]) 
            self.mlp_2 = torch.nn.Sequential(mlp[-1])   
        else:
            self.mlp = torch.nn.Sequential(*mlp)

    def forward(self, x, node_level=None):
        if self.use_multi_head:
            x = self.mlp_1(x)
            x = select_corresponding_level(x, node_level, self.num_levels)
            return self.mlp_2(x)
        else:
            return self.mlp(x)

class TensorReadOut(torch.nn.Module):
    def __init__(
        self,
        l: int,
        in_dim: int,
        hidden_dim: List[int],
        out_dim: int,
        gate: str = "silu",
        num_levels: int = 1,
        use_multi_head: bool = False,
    ) -> None:
        super().__init__()

        self.l = l
        self.num_levels = num_levels
        self.use_multi_head = (use_multi_head) and (num_levels > 1) and len(hidden_dim) > 0
        if self.use_multi_head:
            assert len(hidden_dim) == 1, 'For multihead training, cfg.model.config.readout_emlp.hidden_dim must be only one neuron'

        self.linears_1 = torch.nn.ModuleList()
        prev_dim = in_dim
        for h_dim in hidden_dim:
            self.linears_1.append(
                Linear(
                    in_dim=prev_dim,
                    out_dim=h_dim,
                    bias=False,
                    l=self.l,
                )
            )
            prev_dim = h_dim

        self.gates = torch.nn.ModuleList()
        for dim in hidden_dim:
            self.gates.append(NormGate[gate](self.l, dim))
            
        self.linear_2= Linear(
            in_dim=prev_dim,
            out_dim=out_dim,
            bias=False,
            l=self.l,
        )

    def forward(self, t: torch.Tensor, node_level: torch.Tensor) -> torch.Tensor:
        for idx, linear in enumerate(self.linears_1):
            t = self.gates[idx](linear(t))
        if self.use_multi_head:
            t = select_corresponding_level(t, node_level, self.num_levels)
        return self.linear_2(t)


def build_scalar_readout(
    in_dim: int,
    hidden_dim: List[int],
    act: int,
    bias: bool,
    num_levels: int,
    use_multi_head: bool,
    num_layers: int,
    use_all_layer: bool,
):
    readouts = []
    for idx in range(num_layers):
        if idx == (num_layers-1):
            if use_multi_head:
                hidden_dim_ = [d * num_levels for d in hidden_dim]
                out_dim_ = num_levels
            else:
                hidden_dim_ = hidden_dim
                out_dim_ = 1
            act_ = act
            use_multi_head_ = use_multi_head
            forward_weight_init = False
        else:
            hidden_dim_ = []
            out_dim_ = 1
            act_ = None
            use_multi_head_ = False
            forward_weight_init = True

        readouts.append(
            ScalarReadOut(
                in_dim=in_dim,
                out_dim=out_dim_,
                hidden_dim=hidden_dim_,
                act=act_,
                bias=bias,
                forward_weight_init=forward_weight_init,
                num_levels=num_levels,
                use_multi_head=use_multi_head_,
            )
        )
    if use_all_layer:
        return torch.nn.ModuleList(readouts)
    else:
        return torch.nn.ModuleList([readouts[-1]])


def build_tensor_readout(
    l: int,
    in_dim: int,
    hidden_dim: int,
    gate: int,
    num_levels: int,
    use_multi_head: bool,
    num_layers: int,
    use_all_layer: bool,
):
    readouts = []
    for idx in range(num_layers):
        if idx == num_layers - 1:
            if use_multi_head:
                hidden_dim_ = [d * num_levels for d in hidden_dim]
                out_dim_ = num_levels
            else:
                hidden_dim_ = hidden_dim
                out_dim_ = 1
            gate_ = gate
            use_multi_head_ = use_multi_head
        else:
            hidden_dim_ = []
            out_dim_ = 1
            gate_ = None
            use_multi_head_ = False

        readouts.append(
            TensorReadOut(
                l=l,
                in_dim=in_dim,
                hidden_dim=hidden_dim_,
                out_dim=out_dim_,
                gate=gate_,
                num_levels=num_levels,
                use_multi_head=use_multi_head_,
            )
        ) 
    if use_all_layer:
        return torch.nn.ModuleList(readouts)
    else:
        return torch.nn.ModuleList([readouts[-1]])




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

from typing import Dict, Type

import torch

from .ictd import ICTD


class DirectPolarizability(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        PS, DS, CS, SS = ICTD(2)
        self.register_buffer(
            "one_e", 
            SS[2].view(3, 3).to(dtype=torch.get_default_dtype()),
        )
        del PS, DS, CS, SS

    def forward(self, t0: torch.Tensor, t2: torch.Tensor) -> torch.Tensor:
        t0 = torch.einsum('b, ij -> bij', t0, self.one_e)
        return t0 + t2
    
class DirectVirials(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        PS, DS, CS, SS = ICTD(2)
        self.register_buffer(
            "one_e", 
            SS[2].view(3, 3).to(dtype=torch.get_default_dtype()),
        )
        del PS, DS, CS, SS

    def forward(self, t0: torch.Tensor, t2: torch.Tensor) -> torch.Tensor:
        t0 = torch.einsum('b, ij -> bij', t0, self.one_e)
        return t0 + t2

class DirectHessians(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        PS, DS, CS, SS = ICTD(2)
        self.register_buffer(
            "one_e", 
            SS[2].view(3, 3).to(dtype=torch.get_default_dtype()),
        )
        del PS, DS, CS, SS

    def forward(self, t0: torch.Tensor, t2: torch.Tensor) -> torch.Tensor:
        t0 = torch.einsum('b, ij -> bij', t0, self.one_e)
        return t0 + t2
    
PropertyBasisChange: Dict[str, Type[torch.nn.Module]] = {
    'direct_virials': DirectVirials,
    'direct_polarizability': DirectPolarizability,
    'direct_hessians': DirectHessians,
}






import sys
import torch
import torch_sim as ts
from tace.interface.torchsim import TACETorchSimCalc

# Structure
from ase.build import bulk
atoms = bulk("Cu", "fcc", a=3.58, cubic=True).repeat((2, 2, 2))
atomsList = [atoms] * 3

# TACETorchSimCalc
# model = sys.argv[1]
model = '/home/xuzemin/TACE-OAM-L-R-v1.ckpt'
dtype = torch.float32
device = torch.device("cuda")
torchSimModel = TACETorchSimCalc(model, dtype=dtype, device=device, compute_stress=False)


# run them all simultaneously with batching
trajectory_files = [f"Cu_traj_{i}.h5md" for i in range(len(atomsList))]
final_state = ts.integrate(
    system=atomsList,
    model=torchSimModel,
    n_steps=50,
    timestep=0.002,
    temperature=1000,
    integrator=ts.Integrator.nvt_langevin,
    trajectory_reporter=dict(filenames=trajectory_files, state_frequency=10),
)
final_atoms_list = final_state.to_atoms()

# extract the final energy from the trajectory file
final_energies = []
for filename in trajectory_files:
    with ts.TorchSimTrajectory(filename) as traj:
        final_energies.append(traj.get_array("potential_energy")[-1])

print(final_energies)


# ###############################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
# check: ✔
# ###############################################################################


from typing import Dict, List, Tuple, Optional


import torch
from torch import nn


from .paths import count_irreps, generate_combinations
from .linear import Linear
from .einsum import InterEinsumTC
from ...utils.torch_scatter import scatter_sum


class Contraction(torch.nn.Module):
    """No operator fusion"""
    
    weight_numel: int
    
    def __init__(
        self,
        num_channel: int = 64,
        num_channel_hidden: int = 64,
        lmax_in: int = 3,
        lmax_out: int = 3,
        l1l2: Optional[str] = None,
        filter_combs: Optional[Tuple[int, int, int ,int]] = None,
    ):
        super().__init__()

        old_combs = generate_combinations(
            lmax_in,
            lmax_out,
            lmax_out,
            l1l2=l1l2,
        )

        assert isinstance(filter_combs, List) or filter_combs is None
        filter_combs = filter_combs or []

        new_combs = []
        for comb in old_combs:
            if list(comb) not in filter_combs:
                new_combs.append(comb)

        # ==== ICTP + ICTC ====
        self.tcs = nn.ModuleList()
        for comb in new_combs:
            self.tcs.append(InterEinsumTC(comb))
        l3_count = count_irreps(new_combs, False, False, True, True)

        # === conv_weights slices ===
        self.ws_slices = []
        start = 0
        for _ in new_combs:
            stop = start + num_channel
            self.ws_slices.append(slice(start, stop))
            start = stop

        # === linear ===
        self.linear_downs = nn.ModuleList(
            [
                Linear(
                    num_channel * count,
                    num_channel_hidden,
                    bias=False,
                    l=l3,
                    in_channel=num_channel,
                    out_channel=num_channel_hidden,
                )
                for l3, count in l3_count.items()
            ]
        )

        self.combs = new_combs
        self.lmax_in = lmax_in
        self.lmax_out = lmax_out
        self.weight_numel = num_channel * len(new_combs)

    def forward(
        self,
        x: Dict[int, torch.Tensor],
        y: Dict[int, torch.Tensor],
        ws: torch.Tensor,
        edge_index: torch.Tensor,
    ) -> Dict[int, torch.Tensor]:
        
        num_nodes = x[0].size(0)
        for l1 in range(self.lmax_in + 1):
            x[l1] = x[l1][edge_index[0]]

        buffer = {l3: [] for l3 in range(self.lmax_out + 1)}
        for i, tc in enumerate(self.tcs):
            l1, l2, l3, _ = self.combs[i]
            out  = tc(x[l1], y[l2])
            w = ws[:, self.ws_slices[i]]
            for _ in range(out.ndim - 2):
                w = w.unsqueeze(-1)
            w_out = w * out
            buffer[l3].append(w_out)
        m_ji = {}
        for i, linear in enumerate(self.linear_downs):
            m_ji[i] = linear(torch.cat(buffer[i], dim=1))
        m_i = {}
        for r in m_ji.keys():
            m_i[r] = scatter_sum(
                src=m_ji[r],
                index=edge_index[1],
                dim=0,
                dim_size=num_nodes,
            )

        return m_i




################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################
"""
This file is about the rotation, init_edge_rot_mat from eSCN.
"""
import torch

# def rotate_cart(T: Tensor, R: Tensor):
#     r = T.ndim - 2 
#     if r > 0:
#         in_1 = 'b' + ''.join(LETTERS[0:2])
#         in_2 = 'bc' + ''.join(LETTERS[2:r+2])
#         in_2 = list(in_2)
#         in_2[-1] = in_1[-1]
#         in_2 = ''.join(in_2)
#         out = 'bc' + in_1[1] + in_2[2:-1]
#         einsum_str = in_1 + ',' + in_2 + '->' + out
#         for _ in range(r):
#             T = torch.einsum(einsum_str, R, T)
#     return T


# def init_edge_rot_mat(edge_vector) -> torch.Tensor:
#     """
#     This function is adapted from eSCN: https://openreview.net/forum?id=QIejMwU0r9
#     Returns:
#         edge_rot_mat (edge, 3, 3).
#     """
#     edge_vec_0 = edge_vector
#     edge_vec_0_distance = torch.sqrt(torch.sum(edge_vec_0**2, dim=1))

#     norm_x = edge_vec_0 / (edge_vec_0_distance.view(-1, 1))

#     edge_vec_2 = torch.rand_like(edge_vec_0) - 0.5
#     edge_vec_2 = edge_vec_2 / (
#         torch.sqrt(torch.sum(edge_vec_2**2, dim=1)).view(-1, 1)
#     )
#     # Create two rotated copys of the random vectors in case the random vector is aligned with norm_x
#     # With two 90 degree rotated vectors, at least one should not be aligned with norm_x
#     edge_vec_2b = edge_vec_2.clone()
#     edge_vec_2b[:, 0] = -edge_vec_2[:, 1]
#     edge_vec_2b[:, 1] = edge_vec_2[:, 0]
#     edge_vec_2c = edge_vec_2.clone()
#     edge_vec_2c[:, 1] = -edge_vec_2[:, 2]
#     edge_vec_2c[:, 2] = edge_vec_2[:, 1]
#     vec_dot_b = torch.abs(torch.sum(edge_vec_2b * norm_x, dim=1)).view(-1, 1)
#     vec_dot_c = torch.abs(torch.sum(edge_vec_2c * norm_x, dim=1)).view(-1, 1)

#     vec_dot = torch.abs(torch.sum(edge_vec_2 * norm_x, dim=1)).view(-1, 1)
#     edge_vec_2 = torch.where(torch.gt(vec_dot, vec_dot_b), edge_vec_2b, edge_vec_2)
#     vec_dot = torch.abs(torch.sum(edge_vec_2 * norm_x, dim=1)).view(-1, 1)
#     edge_vec_2 = torch.where(torch.gt(vec_dot, vec_dot_c), edge_vec_2c, edge_vec_2)

#     vec_dot = torch.abs(torch.sum(edge_vec_2 * norm_x, dim=1))
#     # Check the vectors aren't aligned
#     assert torch.max(vec_dot) < 0.99

#     norm_z = torch.cross(norm_x, edge_vec_2, dim=1)
#     norm_z = norm_z / (torch.sqrt(torch.sum(norm_z**2, dim=1, keepdim=True)))
#     norm_z = norm_z / (torch.sqrt(torch.sum(norm_z**2, dim=1)).view(-1, 1))
#     norm_y = torch.cross(norm_x, norm_z, dim=1)
#     norm_y = norm_y / (torch.sqrt(torch.sum(norm_y**2, dim=1, keepdim=True)))

#     # Construct the 3D rotation matrix
#     norm_x = norm_x.view(-1, 3, 1)
#     norm_y = -norm_y.view(-1, 3, 1)
#     norm_z = norm_z.view(-1, 3, 1)

#     edge_rot_mat_inv = torch.cat([norm_z, norm_x, norm_y], dim=2)
#     edge_rot_mat = torch.transpose(edge_rot_mat_inv, 1, 2)

#     return edge_rot_mat.detach()
 

def init_edge_rot_mat(edge_vector) -> torch.Tensor:
    """
    x
    """
    edge_vec_0 = edge_vector
    edge_vec_0_distance = torch.sqrt(torch.sum(edge_vec_0**2, dim=1))
    norm_x = edge_vec_0 / (edge_vec_0_distance.view(-1, 1)) 

    edge_vec_2 = torch.rand_like(edge_vec_0) - 0.5
    edge_vec_2 = edge_vec_2 / torch.sqrt(torch.sum(edge_vec_2**2, dim=1, keepdim=True))

    edge_vec_2b = edge_vec_2.clone()
    edge_vec_2b[:, 0] = -edge_vec_2[:, 1]
    edge_vec_2b[:, 1] = edge_vec_2[:, 0]
    edge_vec_2c = edge_vec_2.clone()
    edge_vec_2c[:, 1] = -edge_vec_2[:, 2]
    edge_vec_2c[:, 2] = edge_vec_2[:, 1]

    vec_dot_b = torch.abs(torch.sum(edge_vec_2b * norm_x, dim=1, keepdim=True))
    vec_dot_c = torch.abs(torch.sum(edge_vec_2c * norm_x, dim=1, keepdim=True))
    vec_dot = torch.abs(torch.sum(edge_vec_2 * norm_x, dim=1, keepdim=True))

    edge_vec_2 = torch.where(vec_dot > vec_dot_b, edge_vec_2b, edge_vec_2)
    vec_dot = torch.abs(torch.sum(edge_vec_2 * norm_x, dim=1, keepdim=True))
    edge_vec_2 = torch.where(vec_dot > vec_dot_c, edge_vec_2c, edge_vec_2)

    vec_dot = torch.abs(torch.sum(edge_vec_2 * norm_x, dim=1))
    assert torch.max(vec_dot) < 0.99

    norm_z = torch.cross(norm_x, edge_vec_2, dim=1)
    norm_z = norm_z / torch.sqrt(torch.sum(norm_z**2, dim=1, keepdim=True))
    norm_y = torch.cross(norm_z, norm_x, dim=1)
    norm_y = norm_y / torch.sqrt(torch.sum(norm_y**2, dim=1, keepdim=True))

    norm_x = norm_x.view(-1, 3, 1)
    norm_y = norm_y.view(-1, 3, 1)
    norm_z = norm_z.view(-1, 3, 1)

    edge_rot_mat_inv = torch.cat([norm_x, norm_y, norm_z], dim=2)
    edge_rot_mat = torch.transpose(edge_rot_mat_inv, 1, 2)

    return edge_rot_mat.detach()



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import copy
import argparse
import time
from typing import Dict


import ase.io
from tqdm import tqdm
import torch
from torch import Tensor
from torch_geometric.loader import DataLoader
from e3nn.util.jit import compile


from ..lightning import load_tace
from ..dataset.element import TorchElement
from ..dataset.graph import from_atoms
from ..utils.metrics import build_metrics, update_metrics
from ..utils._global import DTYPE
from ..utils.utils import num_params
from ..dataset.quantity import KeySpecification, update_keyspec_from_kwargs, PROPERTY
from ..dataset.read import check_keys

def parse_args():
    parser = argparse.ArgumentParser(
        description="Predict properties for structures using a trained model."
        "You could also print test metrics or predict property not in your training data "
        "(such as training with energy only and wishing to predict forces)"
    )
    parser.add_argument("-i", "--input", type=str, required=True, help="Path to ASE-readable input file")
    parser.add_argument("-o", "--output", type=str, default="predict.xyz",help="Path to ASE-writeable output file")
    parser.add_argument("-m", "--model", type=str, required=True, help="Path to model checkpoint (.ckpt or .pt or .pth)")
    parser.add_argument("-t", "--test", type=int, default=0, help="print test set metric")
    parser.add_argument("-e", "--ema", type=int, default=1, help="Use EMA parameters during evaluation")
    parser.add_argument("-b", "--batch_size", type=int, default=16, help="Batch size for inference")
    parser.add_argument("--dtype", type=str, default="float32", choices=["float32", "float64"], help="Tensor precision")
    parser.add_argument("--device", type=str, default="cuda", choices=["cpu", "cuda"], help="Device for inference")
    parser.add_argument("--nl_backend", type=str, default="matscipy", choices=["ase", "vesin", "matscipy"], help="nl_backend")
    # parser.add_argument("-c", "--compile", type=int, default=0, help="Compile to jit-model, not support know")

    # Keys for properties, if need print test metrics
    for k, v in PROPERTY.items():
        if v['enable_prediction'] or v['enable_embedding']:
            parser.add_argument(f"--{k}_key", type=str, default=f"{k}")

    return parser.parse_args()

def main():
    args = parse_args()
    atomsList = ase.io.read(args.input, index=":")
    atoms_list_copy = copy.deepcopy(atomsList)
    key_spec = KeySpecification()
    update_keyspec_from_kwargs(key_spec, vars(args))
    # avoid conflict with ase 
    check = False
    if args.test == 1:
        check = True

    model =load_tace(args.model, args.device, strict=True, use_ema=args.ema)
    max_neighbors = model.max_neighbors.item() if hasattr(model, "max_neighbors") else None
    cutoff = model.readout_fn.cutoff.item()
    atomic_numbers = model.readout_fn.atomic_numbers.tolist()
    device = args.device
    model_dtype = model.readout_fn.cutoff.dtype
    args_dtype = DTYPE[args.dtype]


    target_property = list(set(model.target_property))
    embedding_property = getattr(model, "embedding_property", [])

    model.eval().to(device)
    print(f"Number of parameters: {num_params(model)}")
    if args_dtype != model_dtype:
        print(f"[Warning] Model dtype {(model_dtype)} does not match args.dtype {(args_dtype)}. Forcing dtype to {args_dtype}")
    torch.set_default_dtype(args_dtype)
    model.to(args_dtype)

    # Build metrics
    if args.test == 1:
        metrics: Dict = build_metrics("test", target_property)
        for m in metrics.values():
            m.to(device)

    # Dataset
    element = TorchElement([int(z) for z in atomic_numbers])
    dataset = [
        from_atoms(
            element,
            atoms,
            cutoff,
            max_neighbors="inf" if max_neighbors is None else max_neighbors,
            keyspec=key_spec,
            target_property=target_property,
            embedding_property=embedding_property,
            universal_embedding=getattr(model, 'universal_embedding', {}),
            neighborlist_backend=args.nl_backend,
        )
        for atoms in check_keys(atomsList, target_property, key_spec, embedding_property, check)
    ]
    dataloader = DataLoader(dataset, batch_size=args.batch_size, shuffle=False)

    preds, ptrs = [], []
    start_time = time.time()

    # Run inference
    for batch in tqdm(dataloader):
        batch.to(device)
        for p in target_property:
            for requires_grad_p in PROPERTY[p]['requires_grad_with']:
                batch[requires_grad_p].requires_grad_(True)

        with torch.enable_grad():
            outputs = model(batch)
            if args.test == 1:
                update_metrics(metrics, "test", outputs, batch, target_property)

            preds.append(
                {
                    k: v.detach().cpu() if isinstance(v, Tensor) else None
                    for k, v in outputs.items()
                }
            )
            ptrs.append(batch.ptr.cpu())

    print(f"Inference finished in {time.time() - start_time:.2f} s "
          f"(batch_size={args.batch_size})")

    if args.device == "cuda" and torch.cuda.is_available():
        print(f"Device name: {torch.cuda.get_device_name(torch.cuda.current_device())}")
    else:
        print(f"Device name: {device.upper()}")

    if args.test == 1:  
        # Print metrics
        print("\n====== Evaluation Metrics ======")
        for k, v in metrics.items():
            print(f"{k}: {v.compute().item():.4f}")

    # Assign predictions back to atoms in the same order as atomsList
    pred_atoms_list = []
    structure_index = 0  # global counter across all batches
    for pred, ptr in zip(preds, ptrs):
        n_structures = len(ptr) - 1
        for i in range(n_structures):
            start, end = ptr[i], ptr[i + 1]
            atoms_i = atoms_list_copy[structure_index]
            for p in target_property:
                p_scope = PROPERTY[p]['scope']
                p_rank = PROPERTY[p]['rank']
                if p_scope == 'per-system':
                    if p_rank == 0:
                        atoms_i.info[f"TACE_{p}"] = pred[p][i].item()
                    else:
                        atoms_i.info[f"TACE_{p}"] = pred[p][i].detach().cpu().numpy().reshape(-1)
                elif p_scope == 'per-atom':
                    atoms_i.arrays[f"TACE_{p}"] = pred[p][start:end].detach().cpu().numpy().reshape(-1, 3**p_rank)
                elif p_scope == 'per-edge':
                    atoms_i.info[f"TACE_{p}"] = pred[p][i].detach().cpu().numpy().reshape(-1)
                else:
                    atoms_i.info[f"TACE_{p}"] = pred[p][i].detach().cpu().numpy().reshape(-1)
            pred_atoms_list.append(atoms_i)
            structure_index += 1  # move to next structure

    ase.io.write(args.output, pred_atoms_list, format="extxyz")
    print(f"Results written to {args.output}")


if __name__ == "__main__":
    main()



################################################################################
# Authors: Zemin Xu
# License: MIT, see LICENSE.md
################################################################################

import logging


from lightning.pytorch.callbacks import Callback
from lightning.pytorch.callbacks import TQDMProgressBar


from .ema import ExponentialMovingAverage


# === Print Metriacs Callback ===
class PrintMetricsCallback(Callback):
    def on_train_epoch_end(self, trainer, pl_module):
        self._print_metrics(trainer, pl_module, "train")

    def on_validation_epoch_end(self, trainer, pl_module):
        self._print_metrics(trainer, pl_module, "val")

    # def on_test_epoch_end(self, trainer, pl_module):
    #     self._print_metrics(trainer, pl_module, "test")

    def _print_metrics(self, trainer, pl_module, prefix):
        current_epoch = pl_module.current_epoch
        total_epoch = trainer.max_epochs - 1
        metrics = getattr(pl_module, f"{prefix}_metrics").compute()
        logging.info(f"[Epoch {current_epoch}/{total_epoch}], the error is 1000 times")
        for name, value in metrics.items():
            logging.info(f"{name}: {value.item():.6f}")


# === EMA Callback ===
class EMACallback(Callback):
    def __init__(self, decay: float = 0.99, use_num_updates: bool = True, device=None):
        self.decay = decay
        self.use_num_updates = use_num_updates
        self.device = device
        self.ema = None

    def on_fit_start(self, trainer, pl_module):
        self.ema = ExponentialMovingAverage(
            [p for p in pl_module.model.parameters()], # for freeze convenience
            # [p for p in pl_module.model.parameters() if p.requires_grad],
            self.decay,
            self.use_num_updates,
        )
        self.ema.to(self.device or pl_module.device)

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        self.ema.update()

    def on_validation_start(self, trainer, pl_module):
        # logging.info("[EMA] Applying EMA weights before valid")
        self.ema.store()
        self.ema.copy_to()

    def on_validation_end(self, trainer, pl_module):
        self.ema.restore()

    def on_test_start(self, trainer, pl_module):
        # logging.info("[EMA] Applying EMA weights before test")
        self.ema.store()
        self.ema.copy_to()

    def on_test_end(self, trainer, pl_module):
        self.ema.restore()

    def on_save_checkpoint(self, trainer, pl_module, checkpoint):
        if self.ema:
            checkpoint["ema_state_dict"] = self.ema.state_dict()

    def on_load_checkpoint(self, trainer, pl_module, checkpoint):
        if self.ema and "ema_state_dict" in checkpoint:
            self.ema.load_state_dict(checkpoint["ema_state_dict"])


# class TACEBar(TQDMProgressBar):
#     def __init__(self):
#         super().__init__()
#         self._current_loss = None

#     def get_metrics(self, trainer, model):
#         metrics = {}
#         lr = trainer.optimizers[0].param_groups[0]["lr"]
#         metrics["lr"] = f"{lr:.6f}"
#         if self._current_loss is not None:
#             metrics["loss"] = f"{self._current_loss:.4f}"

#         return metrics


#     def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
#         super().on_train_batch_end(trainer, pl_module, outputs, batch, batch_idx)
        
#         if isinstance(outputs, dict) and "loss" in outputs:
#             self._current_loss = outputs["loss"].item()

#     def on_validation_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0):
#         super().on_validation_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)

#         if isinstance(outputs, dict) and "loss" in outputs:
#             self._current_loss = outputs["loss"].item()

